#!/usr/bin/env python
#
# Copyright 2018-2019 Axel Huebl
#
# License: CC-BY-SA 4.0 International
#
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import matplotlib.cm as cmx
from picongpu.plugins.data import EnergyHistogramData
import numpy as np
from scipy.optimize import curve_fit
import json
import pickle
from collections import defaultdict
from collections.abc import Iterable


# a0=16  # 16, 30, 42

# ...k: ...fs after peak on target
step={
     8: 90000,
    16: 75000,
    23: 70000,
    30: 60000,
    42: 45000
}
# sfilter='fwPinhole'    # 'fwRCF', 'fwPinhole', 'all', 'originFront', 'originRear'
geom='mixed_30nc'  # mixed_30nc separ_30nc  separ_300nc separ_1000nc separ_1700nc

# integrate lower range [MeV]
uboundMeV = 1.0
# cut-off for counts for maximum energy
yrange = [2.e1, 1.e7]
# plot ene limits [MeV]
ene_range = {
     8: 20.,
    16: 50.,
    23: 75.,
    30: 100.,
    42: 200.
}

model = "sim_ePS_avg_rear"  # average n_e and T_e over all d: second half of target
#model = "sim_ePS_db"       # one n_e and T_e per d
ne_hot_ratio_db = {}
Te_hot_sim_db = {}

# important aspect: n0 is only an upper estimate for rear ne!
#   absorption factor for T_hot electrons on front:
#   matched ne for n^hot_e (in rear) for matching T^pond_e
ne_hot_ratio_db["Wilks"] = {
     8: 0.185,  #  2.4 MeV
    16: 0.185,  #  5.3 MeV
    23: 0.21,   #  7.8 MeV
    30: 0.215,  # 10.3 MeV
    42: 0.30,   # 14.7 MeV
}
#   matched ne for n^hot_e (in rear) for matching T^Beg_e
ne_hot_ratio_db["Beg"] = {
     8: 0.33,  # 0.96 MeV
    16: 0.40,  # 1.52 MeV
    23: 0.54,  # 1.94 MeV
    30: 0.60,  # 2.31 MeV
    42: 1.00,  # 2.90 MeV
}
#   matched ne for n^hot_e (in rear) for matching T^Kluge_e
ne_hot_ratio_db["Kluge"] = {
     8: 0.265,  # 1.34 MeV
    16: 0.285,  # 2.58 MeV
    23: 0.35,   # 3.57 MeV
    30: 0.37,   # 4.52 MeV
    42: 0.55,   # 6.07 MeV
}

#   measured in simulation
#   (along onLaserAxis, avg with cutoff correction, rear half of target)
ne_hot_ratio_db["sim_ePS_avg_rear"] = {
     8: 0.483, # includes bulk
  #  8: 0.061,
    16: 0.312,
    23: 0.440,
    30: 0.407,
    42: 0.568,
}
Te_hot_sim_db["sim_ePS_avg_rear"] = {  # MeV
     8: 0.35, # includes bulk
  #  8: 1.34,
    16: 2.17,
    23: 3.52,
    30: 4.52,
    42: 8.64,
}

# see https://stackoverflow.com/a/8702435/2719194
#     https://stackoverflow.com/a/26497091/2719194
def nested_dict(): return defaultdict(nested_dict)

#   measured in simulation (electron phase space) and stored in db
Te_ne_db_name = geom + '/ps_ypy/e_Te' #'_onLaserAxisNarrow'
Te_ne_db_step = 10000
Te_ne_db = nested_dict()
try:
    # restore other geometries if already there
    with open(Te_ne_db_name + '.p', 'rb') as fp:
        Te_ne_db = pickle.load(fp)
except (IOError, EOFError):
    pass

step_peak_target = 8190  # step 15'000 = 53fs after peak @8190
dt = 7.84995e-18  # s

# markers for plots
species_markers = {
    'H': { 'marker': 'o', 'color': 'grey' },
    'D': { 'marker': 'v', 'color': 'black' },
}

# relevant dratios for plotted comparison of spectras
#   select lines that show relevant changes
lines_select = [
    "0.00", "0.20", "0.50", "0.70", "0.80", "0.90", "0.98", "1.00"
]

def merge_bins(counts, bins, num_merge):
    """
    Merge num_merge bins into one to smooth histograms.
    Normalize count by same factor.
    """
    new_counts = np.zeros(counts.shape[0] // int(num_merge), dtype=counts.dtype)
    new_bins = np.zeros(bins.shape[0] // int(num_merge), dtype=bins.dtype)

    new_bins[:] = bins[::num_merge]
    for i in np.arange(num_merge):
        new_counts[:] += counts[i::num_merge]

    new_counts = new_counts / float(num_merge)

    return new_counts, new_bins

def getData(species, sfilter, a0, step, geom, dratio, suffix=""):
    """..."""
    eneHist = EnergyHistogramData("a0_" + str(a0) + "/2D_" + geom + "_dratio" + dratio + suffix)
    counts, bins = eneHist.get(species, sfilter, step)
    bins = bins * 1.e-3  # keV -> MeV
    if species == 'D':
        bins = bins * 0.5

    # smooth histogram by merging neighboring bins
    counts, bins = merge_bins(counts, bins, 8)

    # upper energy range
    max_ene = np.max(bins)
    if species == 'D':
        max_ene *= 2.0

    # calc integrated charge on stack
    firstBin = np.argmax(bins >= uboundMeV)
    sum_counts = np.sum(counts[firstBin:])
    # print("[SUM >{0}MeV, {1}] ".format(uboundMeV, species) + geom + " " + dratio + suffix + ": {0}".format(sum_counts))

    # calc cutoff energy (first time > y-minimum from right)
    cutoffBin = len(counts) - np.argmax(counts[::-1]>=yrange[0]) - 1
    cutoffEne = bins[cutoffBin]
    # print("[cut-off Ene, {1}] ".format(uboundMeV, species) + geom + " " + dratio + suffix + ": {0}".format(cutoffEne))

    return counts, bins, max_ene, sum_counts, cutoffEne

def plot_name(geom, a0, species, sfilter, step):
    if type(species) is list:
        species_name = "-".join(species)
    else:
        species_name = species
    return (
        geom + "/energySpectra/a0_" + str(a0) +
        "_" + species_name + "_" + sfilter + "_" +
        geom + "_" + str(step) + "_ene")

def plotAndSave(species, sfilter, a0, step, geom, dratios, suffix=""):
    """..."""
    f = plt.figure()
    ax = f.gca()
    for dratio in dratios:
        counts, bins, max_ene, _, _= \
            getData(species, sfilter, a0, step, geom, dratio, suffix)
        ax.plot(bins, counts,
            label=(geom + " " + dratio + suffix))

    ax.set_xlabel("Ene / A  [MeV]")
    # ax.set_title("{0} ({1}) @ t = {2}".format(species, sfilter, step))
    ax.set_yscale('log')
    # ax.set_xlim([0., max_ene])
    ax.set_xlim([0., ene_range[a0]])
    ax.set_ylim(yrange)
    ax.legend()

    f.savefig(
        plot_name(geom, a0, species, sfilter, step) + ".png",
        dpi=300)
    plt.close(f)

def plotAndSave_together(species, sfilter, a0, step, geom, dratios, suffix=""):
    """..."""
    f = plt.figure(figsize=[5, 4])
    # left, bottom, W, H
    axD = f.add_axes([0.15, 0.13, 0.8, 0.4])
    axH = f.add_axes([0.15, 0.53, 0.8, 0.4], sharex=axD)

    axCutoff = f.add_axes([0.55, 0.25, 0.35, 0.25])

    cm_plasma_r = plt.get_cmap('plasma_r') 
    cm_norm  = colors.Normalize(vmin=0, vmax=1.0)  # values of d
    lines_color_map = cmx.ScalarMappable(norm=cm_norm, cmap=cm_plasma_r)

    D_cutoffEne = {}
    H_cutoffEne = {}
    H_sum_counts = {}
    D_sum_counts = {}
    H_0_counts = []
    H_0_bins = []
    for dratio in dratios:
        d = float(dratio)
        # d_stretch = d
        d_stretch = d**3 + 0.1*(1.-d) - 0.1*d  # stretch colorspace for selected d
        d_color = lines_color_map.to_rgba(d_stretch)
        if d == 1.0:
            d_color = 'black'

        D_counts, D_bins, D_max_ene, D_sum_counts[d], D_cutoffEne[d] = \
            getData(species[1], sfilter, a0, step, geom, dratio, suffix)
        if dratio in lines_select:
            axD.plot(D_bins, D_counts, color=d_color)
        H_counts, H_bins, H_max_ene, H_sum_counts[d], H_cutoffEne[d] = \
            getData(species[0], sfilter, a0, step, geom, dratio, suffix)
        if d == 0.0:
            H_0_bins = H_bins[:]
            H_0_counts = H_counts[:]
        if dratio in lines_select:
            H_p = axH.plot(H_bins, H_counts,
                color=d_color,
                label=("d=" + dratio + suffix))
            H_modulationBin = np.argmax(H_bins>=D_cutoffEne[d])
            if dratio not in ["1.00", "0.00"]:
                axH.plot((D_cutoffEne[d], D_cutoffEne[d]), (yrange[0], H_counts[H_modulationBin]), linestyle=':', color=H_p[0].get_color())

        # calc integrated charge on filter (pinhole/stack)
        firstBin = np.argmax(H_bins >= D_cutoffEne[d])  # 0.0, uboundMeV
        sum_counts = np.sum(H_counts[firstBin:])
        if d != 1.0:
            print("[SUM >{0:.1f}MeV (D cutoff), {1}] ".format(D_cutoffEne[d], species[0]) + sfilter + " " + dratio + suffix + ": {0:.2f} {1}".format(sum_counts/(1.-d), sum_counts))

    for dratio in dratios:
        d = float(dratio)
        # calc integrated charge on filter (pinhole/stack)
        firstBin = np.argmax(H_0_bins >= D_cutoffEne[d])
        lastBin  = np.argmax(H_0_bins > D_cutoffEne[1.0])
        sum_counts = np.sum(H_0_counts[firstBin:lastBin])
        print("[SUM >{0:.1f}-{1:.1f}MeV, {2}] ".format(D_cutoffEne[d], D_cutoffEne[1.0], species[0]) + sfilter + " " + dratio + suffix + ": {0:.2f}".format(sum_counts))

    # axH.set_title("{0} @ t = {1}".format(geom, step))
    # axD.set_xlabel("Ene / A  [MeV]")
    axD.set_xlabel("$K\ /\ A$  [MeV]")
    f.text(0.02, 0.5,
        'ion count on {}  [arb. u.]'.format(
            sfilter.replace("fwPin", "pin")),
        va='center', rotation='vertical')
    f.text(0.25, 0.85, 'Hydrogen', va='center', rotation='horizontal')
    f.text(0.25, 0.23, 'Deuterium', va='center', rotation='horizontal')
    # axD.set_xlim([0., H_max_ene])
    axD.set_xlim([0., ene_range[a0]])
    axH.tick_params(labelbottom=False)
    axH.set_yscale('log')
    axD.set_yscale('log')
    axH.set_ylim(yrange)
    axD.set_ylim(yrange)
    axD.invert_yaxis()
    axH.legend(ncol=2,
        prop={'size': axH.yaxis.label.get_fontsize() * 0.9})
    f.text(s="$\mathregular{{a}_0}$ = " + "{0}".format(a0),
            color='red',
            ha='right',
            x=0.93, y=0.61)

    # evolution of cutoff energies
    H_d = np.array([float(d) for d in dratios if float(d) != 1.0])
    D_d = np.array([float(d) for d in dratios if float(d) != 0.0])
    H_cutv = np.array([H_cutoffEne[d] for d in sorted(H_cutoffEne.keys()) if float(d) != 1.0])
    D_cutv = np.array([D_cutoffEne[d] for d in sorted(D_cutoffEne.keys()) if float(d) != 0.0])

    # d_to_pos = lambda d: np.searchsorted(H_d[1:], d)

    def get_from_Te_ne_db(name, a0, ds):
        """Return a single entry or an array for a selection of
           Deuterium ratios d (ds)"""
        d_dbs = Te_ne_db[Te_ne_db_step][a0]["d"]
        if isinstance(ds, Iterable):
            vals = np.zeros_like(ds)
            i = 0
            for d in ds:
                d_idx = np.where(d_dbs==d)[0][0]
                vals[i] = Te_ne_db[Te_ne_db_step][a0][name][d_idx]
                i += 1
            return vals
        else:
            d_idx = np.where(d_dbs==ds)[0][0]
            val = Te_ne_db[Te_ne_db_step][a0][name][d_idx]
            return val

    # Density Model Selection [n_0]
    if model == "sim_ePS_db":
        def ne_hot_ratio(a0, ds):
            return get_from_Te_ne_db("ne0", a0, ds)
    else:
        ne_hot_ratio = lambda a0, d=None: ne_hot_ratio_db[model][a0]

    # Temperature Models [MeV]
    T_hot_pond = lambda a0: ( (1.+a0**2/2.)**0.5 - 1. ) * 0.511            # a0 <= 1 (Wilks, 2001)
    T_hot_Beg = lambda a0: 0.469*a0**(2./3.) * 0.511                       # a0 >> 1 (Beg, 1997)
    T_hot_Kluge = lambda a0: ( np.pi*a0/(2.*np.log(4.*a0)) - 1. ) * 0.511  # a0 >> 1 (Kluge, 2011)

    two_temperature_w_correction = 1.0                                     # two temperature mixing
    two_temperature_n_correction = 1.0                                     #   bulk only relevant in
    two_temperature_T_correction = 1.0                                     #   a0=8 and 16 cases
    if model == "Wilks":
        T_hot = lambda a0, d=None: T_hot_pond(a0) * np.ones_like(d)
    elif model == "Beg":
        T_hot = lambda a0, d=None: T_hot_Beg(a0) * np.ones_like(d)
    elif model == "Kluge":
        T_hot = lambda a0, d=None: T_hot_Kluge(a0) * np.ones_like(d)
    # measured in simulation
    elif model == "sim_ePS_db":
        def T_hot(a0, ds):
            return get_from_Te_ne_db("Te", a0, ds)
    else:
        T_hot = lambda a0, d=None: Te_hot_sim_db[model].get(a0) * np.ones_like(d)

    # constants
    kB = 1.38064852e-23  # m^2 kg s^-2 K^-1
    eps0 = 8.854188e-12  # F/m
    qe = 1.60217662e-19  # C
    me = 9.10938356e-31  # kg
    mH = 1.6726219e-27   # kg
    mD = 2. * mH
    ZH = 1.
    ZD = 1.
    AH = 1.              # nucleons
    AD = 2.
    nC = 1.74e27         # m^-3
    n0_nC = lambda d: 30. * ne_hot_ratio(a0, d)  # nC
    MeV2J = 1.60218e-13  # J
    J2MeV = 1. / MeV2J   # MeV / J
    J2K = 1. / kB        # K / J
    K2J = kB             # J / K

    # plasma fundamentals
    omega_pi = lambda n, m, Z: (n*(Z*qe)**2/m/eps0)**0.5
    cs_i = lambda Te, m, Z: (Z*kB*Te/m)**0.5
    Te_cs = lambda cs_i, m, Z: cs_i**2 * m / (Z*kB)

    # shortcuts per ion species
    omega_pH = lambda d: omega_pi((1-d) * n0_nC(d) * nC, mH, ZH)
    omega_pD = lambda d: omega_pi(d * n0_nC(d) * nC, mD, ZD)
    # omega_pe = omega_pi(n0_nC * nC, me, 1.)
    cs_H = lambda Te: cs_i(Te, mH, ZH)
    cs_D = lambda Te: cs_i(Te, mD, ZD)

    # Deuterium cutoff energy model
    t = (step - step_peak_target) * dt
    tauH = t * omega_pH(0.0)  # normed e- time (does NOT scale with d!)
    tauD = t * omega_pD(1.0)  # normed e- time (does NOT scale with d!)
    # taue = t * omega_pe
    tauD = 1.3 * 30.e-15 * omega_pD(1.0)  # Fuchs in NatPhy 2006

    mr = mH/mD
    sqmr = mr**0.5
    fac = lambda Te: ZD / ZH * np.sqrt(ZH * kB * Te / mH) / np.e
    D_cutv_pure = D_cutv[-1]
    D_model_cutoff = lambda a0, d, Te: J2MeV / AD * mD / 2. * (
        np.sqrt(D_cutv_pure * AD * MeV2J * 2. / mD)
        - (1.-d) * fac( Te(a0, d) * MeV2J * J2K ) * (
            (np.e**3 / 2.)**sqmr * (tauD**(2.-2.*sqmr) -1.)/(2.-2.*sqmr)
            - np.e**3 * mr / 2. * np.log(tauD)
        )
    )**2

    # Tikhonchuk (2005): assumes large ratio between ions
    #   - derived for very large d (N>>1)
    #     (small Hydrogen impurities)
    #   - paper describes H "heavy" and "L" light
    #   - we will compare with Ene_Heavy since this is our proclaimed
    #     Deuterium cutoff which modules into the protons
    #   - Ene_Light is the light ion low-energy cutoff
    N_Tikhonchuk = lambda d: ZD / ZH * (d) / (1.-d)
    a_Tikhonchuk = AD * ZD / (AH * ZD)
    D_Tikhonchuk = lambda a0, d, Te: 0.5 * ZD * Te(a0, d) * np.log(
        4. * np.sqrt(2. * a_Tikhonchuk) * N_Tikhonchuk(d) / np.e
    )**2
    H_Tikhonchuk = lambda a0, d, Te: ZH * Te(a0, d) * np.log(
        4. * np.sqrt(2. * a_Tikhonchuk) * N_Tikhonchuk(d) / np.e
    )

    if model == "sim_ePS_db":  # measured n_e and T_e per d
        D_pts = D_d.copy()
        H_pts = H_d.copy()
    else:                      # average n_e and T_e
        D_pts = np.linspace(0.0000001, 0.999999, 200)
        H_pts = np.linspace(0.0000001, 0.9999999999999999, 200)

    dKinDeut = np.array(D_model_cutoff(a0, D_pts, T_hot))

    # Tikhonchuk (2005): assumes large ratio between ions and large d
    # originally with ponderomitive e- scaling, we can take measured T_e
    # to improve its prediction
    #   low-energy cutoff of the lighter species
    axCutoff.plot(
        H_pts,
        H_Tikhonchuk(a0, H_pts, T_hot),
        #H_Tikhonchuk(a0, H_pts, lambda a0, d=None: T_hot_pond(a0) * np.ones_like(d)),
        color=species_markers['H']['color'], linestyle=':')
    #   cutoff energy of the heavier species
    axCutoff.plot(
        D_pts,
        D_Tikhonchuk(a0, D_pts, T_hot) / AD,
        #D_Tikhonchuk(a0, D_pts, lambda a0, d=None: T_hot_pond(a0) * np.ones_like(d)),
        color=species_markers['D']['color'], linestyle=':')

    # new model
    axCutoff.plot(
        D_pts,
        dKinDeut,
        #D_pts[20:],
        #D_model_cutoff(D_pts[20:], 1.9),
        #D_d[:-1],
        #D_model_cutoff(D_d[:-1], *popt),
        color='red', linestyle='-')

    # Mora eq. (22) with T_hot model and effective acceleration time
    #axCutoff.axhline(
    #    2. * T_hot(a0) * \
    axCutoff.plot(
        H_pts,
        2. * T_hot(a0, D_pts) * \
        1.3 * 30.e-15 * omega_pH(0.0),  # Fuchs NatPhy 2006
        color=species_markers['H']['color'],
        linestyle='-')

    axCutoff.plot(
        H_d, H_cutv,
        label='H',
        marker=species_markers['H']['marker'], color=species_markers['H']['color'],
        fillstyle='none', linestyle='none')
    axCutoff.plot(
        D_d, D_cutv,
        label='D',
        marker=species_markers['D']['marker'], color=species_markers['D']['color'],
        fillstyle='none', linestyle='none')
    print("D cutv(d):", D_d, D_cutv)

    axCutoff.set_ylim([0, np.max(H_cutv) * 1.1])

    axCutoff.set_xlabel("Deuterium ratio d")
    # axCutoff.set_ylabel("max. Ene / A  [MeV]",
    axCutoff.set_ylabel(r"$K^\mathrm{ctf}\ /\ A$  [MeV]",
        horizontalalignment='right', y=1.0)
    axCutoff.legend()  # loc='center left')

    f.savefig(
        plot_name(geom, a0, species, sfilter, step) + ".svg",
        dpi=300)
    plt.close(f)

    # save cutoff energies for later use
    save_name = geom + '/energyCutoff/cutoffs_' + sfilter
    save_cutoff = nested_dict()
    try:
        # restore other geometries if already there
        with open(save_name + '.p', 'rb') as fp:
            save_cutoff = pickle.load(fp)
    except (IOError, EOFError):
        pass

    for d in sorted(H_cutoffEne.keys()):
        if float(d) != 1.0:
            save_cutoff[geom][a0][species[0]][d] = H_cutoffEne[d]
    for d in sorted(D_cutoffEne.keys()):
        if float(d) != 0.0:
            save_cutoff[geom][a0][species[1]][d] = D_cutoffEne[d]

    with open(save_name + '.json', 'w') as fp:
        json.dump(save_cutoff, fp, sort_keys=True, indent=4)
    with open(save_name + '.p', 'wb') as fp:
        pickle.dump(save_cutoff, fp)

    # total charge on RCF >=uboundMeV
    fCnt = plt.figure()
    axCnt = fCnt.gca()

    H_cntv = [H_sum_counts[d] for d in sorted(H_sum_counts.keys()) if float(d) != 1.0]
    D_cntv = [D_sum_counts[d] for d in sorted(D_sum_counts.keys()) if float(d) != 1.0]
    axCnt.plot(
        H_d,
        H_cntv,
        label='H', marker='o', color='grey', fillstyle='none', linestyle='none')
    axCnt.plot(
        D_d,
        D_cntv,
        label='D', marker='v', color='black', fillstyle='none', linestyle='none')
    axCnt.set_title("summed ion count above {0} MeV @ t = {1}".format(uboundMeV, step))
    axCnt.set_xlabel("Deuterium ratio d")
    axCnt.set_ylabel('summed ion count on {0}  [arb. u.]'.format(
        sfilter.replace("fwPin", "pin")
    ))

    cnt_name = (geom + "/totalCharge/a0_" + str(a0) +
        "_" + "-".join(species) + "_" + sfilter + "_" +
        geom + "_" + str(step) + "_above" + str(uboundMeV) + "_totalCharge")
    fCnt.savefig(
        cnt_name + ".svg",
        dpi=300)
    plt.close(fCnt)

all_dratios = [
    "0.00", "0.10", "0.20", "0.30", "0.40", "0.50",
    "0.60", "0.70", "0.80", "0.90", "0.98", "1.00"
]

for geom in ['mixed_30nc']:  # 'separ_30nc'
    for a0 in [8, 16, 23, 30, 42]:
        print("a0 =", a0)
        for sfilter in ['fwPinhole', 'originFront', 'originRear', 'fwRCF', 'all']:
            plotAndSave_together(['H', 'D'], sfilter, a0, step[a0], geom,
                                 dratios=all_dratios)
            #for species in ['H', 'D']:
            #    plotAndSave(species, sfilter, a0, step[a0], geom,
            #                dratios=lines_select)
