#!/usr/bin/env python
#
# Copyright 2018-2019 Axel Huebl
#
# License: CC-BY-SA 4.0 International
#
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from picongpu.plugins.data import EnergyHistogramData
import numpy as np
from scipy.optimize import curve_fit
import json
import pickle
from collections import defaultdict


# markers for plots
species_markers = {
    'H': { 'marker': 'o', 'color': '#BBBBBB' },
    'D': { 'marker': 'v', 'color': 'black' },
}

#   measured in simulation
#   (along onLaserAxis, avg with cutoff correction, rear half of target)
ne_hot_ratio_ePS_avg_rear = {
     8: 0.483, # includes bulk
  #  8: 0.061,
    16: 0.312,
    23: 0.440,
    30: 0.407,
    42: 0.568,
}
Te_hot_sim_ePS_avg_rear = {  # MeV
     8: 0.35, # includes bulk
  #  8: 1.34,
    16: 2.17,
    23: 3.52,
    30: 4.52,
    42: 8.64,
}

# see https://stackoverflow.com/a/8702435/2719194
#     https://stackoverflow.com/a/26497091/2719194
def nested_dict(): return defaultdict(nested_dict)

# include 1D Mora scaling with Kluge electron temperature scaling no preplasma
def scalingL0(a0, thickness=2.0):
    """thickness in microns"""
    d = thickness*np.sqrt(2.)
    a0 = a0
    lam = 0.8 # wavelength in um
    theta = 12. # e- opening angle in degrees
    w0 = 1.5 # radius focus spot in um
    W = w0 + np.tan(theta/180.*np.pi)*d
    eta = 0.6  # absorption fudge factor
    tau_p = 30.  # pulse length in fs
    T_PRL_L0 = 0.511 * (np.pi / (2.*(2.3097*a0**(-0.789)))-1.)
    t_PRL_L0 = 1./a0 * np.sqrt(2.*np.exp(1.)*1836.*T_PRL_L0/(eta*(w0/W)**2.))/(2.*np.pi) * 3.33*lam
    e_PRL_L0 = 2*T_PRL_L0*(np.log(tau_p/t_PRL_L0 + np.sqrt(1.+tau_p**2./(t_PRL_L0)**2.)))**2
    return e_PRL_L0

# with preplasma
def scalingL_be_lam_2(a0, thickness=2.0):
    d = thickness
    lam = 0.8  # wavelength in um
    theta = 12.  # e- opening angle in degrees
    w0 = 1.5  # radius focus spot in um
    W = w0 + np.tan(theta/180.*np.pi)*d
    eta = 0.6  # absorption fudge factor
    tau_p = 30.  # pulse length in fs
    T_PRL_L_be_lam_2 = (2.87843*10.**(-10.)*a0**6.-4.33665*10**(-8.)*a0**5.+2.*10**(-6.)*a0**4.+2.*10.**(-5.)*a0**3.-0.005*a0**2.+0.3145*a0-0.0843)*0.511
    t_PRL_L_be_lam_2 = 1./a0 * np.sqrt(2.*np.exp(1.)*1836.*T_PRL_L_be_lam_2/(eta*(w0/W)**2.))/(2.*np.pi) * 3.33*lam
    e_PRL_L_be_lam_2 = 2*T_PRL_L_be_lam_2*(np.log(tau_p/t_PRL_L_be_lam_2 + np.sqrt(1.+tau_p**2./(t_PRL_L_be_lam_2)**2.)))**2
    return e_PRL_L_be_lam_2


def schreiber_ene_max(T_e, Z=1, tau=30.e-15):
    """
    tau in s
    T_e in K
    """
    k_B = 1.38065e-23  # J / K
    J_to_MeV = 6.2415e12

    ene_0 = Z * k_B * T_e

    return J_to_MeV * 2 * ene_0 * np.ln(2. * tau)**2

def getCutoffData(sfilter, geom):
    save_name = geom + '/energyCutoff/cutoffs_' + sfilter
    with open(save_name + '.p', 'rb') as fp:
        save_cutoff = pickle.load(fp)
    return save_cutoff

def plotCutoffAll(species=['H'], sfilter='all', geom='mixed_30nc'):
    save_cutoff = getCutoffData(sfilter, geom)

    save_front = None
    save_rear = None
    if not sfilter.startswith('origin'):
        save_front = getCutoffData('originFront', geom)
        save_rear = getCutoffData('originRear', geom)

    fCutoff = plt.figure(figsize=[4, 3])

    # left, bottom, W, H
    axCutoff = fCutoff.add_axes([0.15, 0.45, 0.65, 0.50])
    cax = fCutoff.add_axes([0.83, 0.15, 0.03, 0.80])
    axHshift = fCutoff.add_axes([0.15, 0.15, 0.65, 0.25], sharex=axCutoff)

    axHshift.axhline(y=1.0,
            zorder=0,
            linestyle=':',
            linewidth=0.7,
            color='grey')

    #for cEene in [1., 10., 100.]:
    #    axCutoff.axhline(y=cEene,
    #            zorder=0,
    #            linestyle=':',
    #            linewidth=0.7,
    #            color='grey')

    cbar = None

    for a0 in save_cutoff[geom]:
        a0f = float(a0)
        #print(save_cutoff[geom][a0])
        #for d in save_cutoff[geom][a0]:
        #    print("d={}, H='".format(save_cutoff[geom][a0][species][d]))

        for s in species:
            # mix all Deuterium ratios (d) together
            d_sorted = sorted(save_cutoff[geom][a0][s].keys())
            cutoffv = [save_cutoff[geom][a0][s][d] for d in d_sorted]

            edgecolor = species_markers[s]['color']
            if save_front is not None:
                cutoff_front = [save_front[geom][a0][s][d] for d in d_sorted]
                cutoff_rear = [save_rear[geom][a0][s][d] for d in d_sorted]
                print(a0, s, sfilter, d_sorted[0], cutoff_front[0], cutoff_rear[0])
                edgecolor = [
                    species_markers[s]['color'] if f < r
                    else 'red'
                    for (f, r) in zip(cutoff_front, cutoff_rear)]

            # d_colors = plt.rcParams['axes.prop_cycle'].by_key()['color']

            #axCutoff.plot(
            #    np.ones_like(cutoffv) * a0f,
            #    cutoffv,
            #    label=s,
            #    marker=species_markers[s]['marker'],
            #    color=species_markers[s]['color'],
            #    fillstyle='none', linestyle='none')

            im = axCutoff.scatter(
                np.ones_like(cutoffv)[::-1] * a0f,
                cutoffv[::-1],
                c = d_sorted[::-1],
                vmin=0., vmax=1.0,
                s=40,
                alpha=0.7,
                cmap='plasma_r',
                linewidths=0.7,
                zorder=1,
                label=s,
                marker=species_markers[s]['marker'],
                edgecolor=edgecolor)
            if cbar is None:
                cbar = plt.colorbar(im, cax=cax)
                cbar.set_label(r'Deuterium ratio $d$')
                cbar.solids.set_edgecolor("face")

            if s == 'H':
                rel_shift = np.array(cutoffv) / cutoffv[0]
                print("{0}: H ({1}) {2}% - {3}%".format(
                    a0, sfilter, np.min(rel_shift), np.max(rel_shift)))
                axHshift.scatter(
                    np.ones_like(cutoffv)[::-1] * a0f,
                    rel_shift[::-1],
                    c = d_sorted[::-1],
                    vmin=0., vmax=1.0,
                    s=40,
                    alpha=0.7,
                    cmap='plasma_r',
                    linewidths=0.7,
                    zorder=1,
                    label=s,
                    marker=species_markers[s]['marker'],
                    edgecolor=edgecolor)

            min_d = d_sorted[0]
            max_d = d_sorted[-1]

            min_v = save_cutoff[geom][a0][s][min_d]
            max_v = save_cutoff[geom][a0][s][max_d]
            print(a0f, min_v, max_v)

            #axCutoff.annotate('{}'.format(min_d),
            #    xy=(a0f, min_v),
            #    xytext=(a0f - 5.5, min_v),
            #    color=species_markers[s]['color'],
            #    arrowprops=dict(
            #        color=species_markers[s]['color'],
            #        # edgecolor='white',
            #        shrink=0.05,
            #        width=2, headwidth=4),
            #)
            #axCutoff.annotate('{}'.format(max_d),
            #    xy=(a0f, max_v),
            #    xytext=(a0f + 3., max_v),
            #    color=species_markers[s]['color'],
            #    arrowprops=dict(
            #        color=species_markers[s]['color'],
            #        # edgecolor='white',
            #        shrink=0.2,
            #        width=2, headwidth=4),
            #)

    # theory
    a = np.linspace(4., 47., 100)
    kluge_pure = scalingL0(a)
    kluge_preplasma = scalingL_be_lam_2(a)
    # axCutoff.plot(a, kluge_pure, zorder=0, color='black', ls='--')
    # axCutoff.plot(a, kluge_preplasma, zorder=0)
    
    # Temperature Models [MeV]
    T_hot_pond = lambda a0: ( (1.+a0**2/2.)**0.5 - 1. ) * 0.511            # a0 <= 1 (Wilks, 2001)
    T_hot_Beg = lambda a0: 0.469*a0**(2./3.) * 0.511                       # a0 >> 1 (Beg, 1997)
    T_hot_Kluge = lambda a0: ( np.pi*a0/(2.*np.log(4.*a0)) - 1. ) * 0.511  # a0 >> 1 (Kluge, 2011)

    # Mora (2003) eq. (22) with <T_e>, n_e measured and
    # effective acceleration time (Fuchs, NatPhy 2006)
    eps0 = 8.854188e-12  # F/m
    qe = 1.60217662e-19  # C
    mH = 1.6726219e-27   # kg
    mD = 2. * mH
    ZH = 1.
    ZD = 1.
    nC = 1.74e27         # m^-3
    ne_hot_ratio = lambda a0: ne_hot_ratio_ePS_avg_rear[a0]  # n0
    n0_nC = lambda a0: 30. * ne_hot_ratio(a0)  # nC

    omega_pi = lambda n, m, Z: (n*(Z*qe)**2/m/eps0)**0.5
    #omega_pH = lambda a0: omega_pi(n0_nC(a0) * nC, mH, ZH)
    omega_pH = lambda a0, ne_n0=1.0: omega_pi(30. * ne_n0 * nC, mH, ZH)

    #mora_H = lambda a0: 2. * ZH * Te_hot_sim_ePS_avg_rear[a0] * 1.3 * 30.e-15 * omega_pH(a0)
    mora_H = lambda a0, ne_n0=1.0: 2. * ZH * T_hot_Kluge(a0) * 1.3 * 30.e-15 * omega_pH(a0, ne_n0)

    #a0 = np.array([8., 16., 23., 30., 42.])
    #mora_H_v = np.array([mora_H(a) for a in a0])
    #axCutoff.plot(
    #    a0,
    #    mora_H_v,
    #    zorder=0, color='black',
    #    ls='--',
    #    # ls='None',
    #    #marker='s',
    #    #markerfacecolor='None'
    #)
    axCutoff.plot(
        a,
        mora_H(a, 0.35),
        zorder=0, color='black',
        ls='--'
    )

    #axCutoff.set_xlabel(r'$a_0$')
    #axCutoff.set_ylabel(r'max. Ene / A  [MeV]')
    axCutoff.set_ylabel(r'$K^\mathrm{ctf}\ /\ A$  [MeV]')
    axCutoff.set_yscale('log')
    axCutoff.set_xlim([3., 48.])
    # axCutoff.set_ylim([0., 170.])
    axCutoff.set_ylim([7.e-1, 240.])

    axHshift.set_xlabel(r'$a_0$')

    axCutoff.text(s="a)",
            fontweight='bold',
            color='black',
            ha='left',
            x=4.0, y=1.0)
    axHshift.text(s="b)",
            fontweight='bold',
            color='black',
            ha='left',
            x=4.0, y=0.75)

    legend_species = [
        Line2D([0], [0], marker=species_markers['H']['marker'], ls='',
               color=species_markers['H']['color'], label='H',
               markersize=6, markerfacecolor='white'),
        Line2D([0], [0], marker=species_markers['D']['marker'], ls='',
               color=species_markers['D']['color'], label='D',
               markersize=6, markerfacecolor='white')
    ]
    #legend_kluge = [
    #    Line2D([0], [0], color='black', lw=1, ls='--', label='Kluge')]
    legend_mora = [
        Line2D([0], [0], color='black', lw=1, ls='--', label='Mora')]
    legend_front = [
        Line2D([0], [0], marker=species_markers['H']['marker'], ls='',
               color='red', label='front origin',
               markersize=6, markerfacecolor='white')]

    leg_s = axCutoff.legend(handles=legend_species, loc='upper left')
    # leg_k = axCutoff.legend(handles=legend_kluge + legend_front,
    leg_k = axCutoff.legend(handles=
        legend_mora +
        legend_front,
        loc='lower right')
    #leg_f = axHshift.legend(handles=legend_front, loc='upper center',
    #    prop={'size': axCutoff.yaxis.label.get_fontsize() * 0.8})
    axCutoff.add_artist(leg_s)
    axCutoff.tick_params(labelbottom=False)

    axHshift.set_ylabel('ratio to pure H',
        size=axCutoff.yaxis.label.get_fontsize() * 0.8)
    axHshift.set_ylim([0.70, 1.25])
    axHshift.set_yticks([0.8, 1.0, 1.2])

    cutoff_name = (geom + "/energyCutoff/" +
        "-".join(species) + "_" + sfilter + "_" +
        geom + "_energyCutoff")
    fCutoff.savefig(
        cutoff_name + ".svg",
        dpi=300)
    plt.close(fCutoff)


for geom in ['mixed_30nc']:  # 'separ_30nc'
    for sfilter in ['fwPinhole']: #, 'fwRCF', 'all', 'originFront', 'originRear']:
        plotCutoffAll(['H', 'D'], sfilter, geom=geom)
        # plotCutoffAll(['H'], sfilter, geom=geom)
        # plotCutoffAll(['D'], sfilter, geom=geom)
