#!/usr/bin/env python
#
# Copyright 2018-2019 Axel Huebl
#
# License: CC-BY-SA 4.0 International
#
import matplotlib as mpl
#mpl.use('Agg') # non-interactive backend
import matplotlib.pyplot as plt
from matplotlib.colors import colorConverter, LogNorm, Normalize
from picongpu.plugins.data import PhaseSpaceData
from scipy.special import lambertw
from scipy.optimize import fsolve, curve_fit
import numpy as np
import pandas as pd
import pickle
from collections import defaultdict
from os import sys, path


lon_mid_y = 4.  # microns
L_x = 7488      # number of cells, transversal box

# selection for analysis of temperature and ene dens
low_dy_mid = -0.0   # microns
hig_dy_mid = +0.9   # microns
mom_cutoff = 1.0    # [beta*gamma] above this we considere e- as "hot"/"laser"/"prompt"
mom_high_cutoff = 99.e9
include_return_currents = False  # calc similar to absorption

step_peak_target = 8190  # 64.3 fs
dt = 3.33e-9 / ( 1.415 * 2.99792458e8 )  # 7.84995e-18  # s
tau = 30.e-15  # s

lambda_0 = 800.e-9  # m
n0 = 30.            # n_c
mu = 1.e6
nc = 1.11485e21 * 1.e6 / (lambda_0 * mu) / (lambda_0 * mu)

q_e = 1.602e-19      # C
MeV2J = 1.60218e-13  # J

# see https://stackoverflow.com/a/8702435/2719194
#     https://stackoverflow.com/a/26497091/2719194
def nested_dict(): return defaultdict(nested_dict)

def getData(sfilter, geom):
    """..."""
    # pickle results
    save_name = geom + '/ps_ypy/e_Te' # + '_' + sfilter
    saved_results = nested_dict()
    try:
        # restore other geometries if already there
        with open(save_name + '.p', 'rb') as fp:
            saved_results = pickle.load(fp)
    except (IOError, EOFError):
        pass

    return saved_results

def getCutoffData(sfilter, geom):
    save_name = geom + '/energyCutoff/cutoffs_' + sfilter
    with open(save_name + '.p', 'rb') as fp:
        save_cutoff = pickle.load(fp)
    return save_cutoff

def plotAndSave(sfilter, a0s, step, geom, dratios):
    """..."""
    saved_results = getData(sfilter, geom)

    for a0 in a0s:
        dratios = saved_results[step][a0]["d"]
        print(dratios)
        
        # ion cutoff data
        save_cutoff = getCutoffData(sfilter='fwPinhole', geom=geom)
        H_d = np.array(sorted(save_cutoff[geom][a0]['H'].keys()))
        H_cutoffv = np.array([save_cutoff[geom][a0]['H'][d] for d in H_d])

        ne = saved_results[step][a0]["ne0"]  # n_0
        Te = saved_results[step][a0]["Te"]   # MeV
        we = saved_results[step][a0]["we"]   # MeV * n_c

        # mean
        ne_avg = saved_results[step][a0]["ne0_avg"]
        Te_avg = saved_results[step][a0]["Te_avg"]
        we_avg = saved_results[step][a0]["we_avg"]

        # standard deviation from mean
        ne_std = saved_results[step][a0]["ne0_std"]
        Te_std = saved_results[step][a0]["Te_std"]
        we_std = saved_results[step][a0]["we_std"]

        # plot dependency on d with theory
        f = plt.figure(figsize=[4, 2.8])

        # left, bottom, W, H
        ax_cutoff = f.add_axes([0.2, 0.15, 0.57, 0.375])
        ax_NeT = ax_cutoff.twinx()
        ax_Ne = f.add_axes([0.2, 0.525, 0.57, 0.375], sharex=ax_cutoff)
        ax_T = ax_Ne.twinx()

        # L_front = [L_front_nm[d] for d in sorted(L_front_nm.keys())]

        MeV2J = 1.60218e-13  # J
        J2MeV = 1. / MeV2J   # MeV / J

        # plot
        #p_cutoff, = ax_cutoff.plot(
        #    dratios,
        #    cutoffv,
        #    color='grey',
        #    marker='x',
        #    label=r'max. Ene / A')
        p_NeT, = ax_NeT.plot(
            dratios,
            we,
            marker='o',
            ls="--",
            color='red',
            label=r'$<w_e>$  [MeV $\cdot$ $n_c$]')
        p_Ne, = ax_Ne.plot(
            dratios,
            ne,
            marker='o',
            ls="--",
            color='blue',
            label=r'$n^\mathrm{laser}_e$  [$n_0$]')
        p_T, = ax_T.plot(
            dratios,
            Te,
            marker='x',
            color='black',
            label=r'$<T_e>$  [MeV]')
        p_cutoff, = ax_cutoff.plot(
            H_d,
            H_cutoffv,
            marker='o',
            # fillstyle='none',
            markerfacecolor='white',
            color='grey',
            label=r'max. Ene (H)\n[MeV]')

        # theoretical predictions for T_e
        # Temperature Models [MeV]
        T_hot_pond = lambda a0: ( (1.+a0**2/2.)**0.5 - 1. ) * 0.511            # a0 <= 1 (Wilks, 2001)
        T_hot_Beg = lambda a0: 0.469*a0**(2./3.) * 0.511                       # a0 >> 1 (Beg, 1997)
        T_hot_Kluge = lambda a0: ( np.pi*a0/(2.*np.log(4.*a0)) - 1. ) * 0.511  # a0 >> 1 (Kluge, 2011)
        ax_T.axhline(
            T_hot_Kluge(a0),
            color="black",
            linestyle=':',
            label=r'Kluge 2011 ($a_0 \gg 1$)')

        # limits
        ax_Ne.set_ylim([0, 1.0])
        ax_T.set_ylim([0, max(T_hot_Kluge(a0), np.max(Te)) * 1.2])
        if a0 == 8:  # comparison of Te with/without bulk
            ax_T.set_ylim([0, 1.8])
        ax_cutoff.set_ylim([0, max(H_cutoffv) * 1.1])
        ax_NeT.set_ylim([np.min(we) * 0.8, np.max(we) * 1.2,])

        # annotate
        #ax_cutoff.legend(handles=[p_cutoff, p_NeT])
        #ax_T.legend(handles=[p_T, p_Ne])

        ax_Ne.tick_params(labelbottom=False)
        ax_NeT.tick_params(labelbottom=False)

        #ax_cutoff.xaxis.set_major_locator(plt.MaxNLocator(5))
        ax_cutoff.set_xticks([0.0, 0.25, 0.5, 0.75, 1.0])

        ax_cutoff.spines['left'].set_color('grey')
        ax_cutoff.yaxis.label.set_color('grey')
        ax_cutoff.tick_params(axis='y', colors='grey')
        ax_Ne.spines['left'].set_color('blue')
        ax_Ne.yaxis.label.set_color('blue')
        ax_Ne.tick_params(axis='y', colors='blue')

        ax_NeT.spines['right'].set_color('red')
        ax_NeT.yaxis.label.set_color('red')
        ax_NeT.tick_params(axis='y', colors='red')
        ax_T.spines['right'].set_color('black')
        ax_T.yaxis.label.set_color('black')
        ax_T.tick_params(axis='y', colors='black')

        ax_cutoff.set_xlabel('Deuterium ratio $d$')
        ax_cutoff.set_ylabel('max. Ene (H)\n[MeV]')
        ax_NeT.set_ylabel('$<w_e>$\n[MeV $\cdot$ $n_c$]')
        ax_T.set_ylabel('$<T_e>$\n[MeV]')
        ax_Ne.set_ylabel('$n_e$\n[$n_0$]')
        #ax_NeT.set_ylabel('$\int_{5\,\mathrm{\mu m}}^{\infty} T\,n_e$\n[arb.u.]')
        #ax_T.set_ylabel('$\int_{5\,\mathrm{\mu m}}^{\infty} T$\n[arb.u.]')
        #ax_Ne.set_ylabel('$\int_{5\,\mathrm{\mu m}}^{\infty} n_e$\n[arb.u.]')

        ax_Ne.set_title(r'$a_0 = {0}$'.format(a0), color='red')

        fname = (geom + "/ps_ypy/a0_" +
            str(a0) + "_" + sfilter + "_" +
            geom +
            "_Te_ne_we_")
        plt.savefig(fname + str(step) + ".svg", dpi=300)
        plt.close()

all_dratios = [
    "0.00", "0.10", "0.20", "0.30", "0.40", "0.50",
    "0.60", "0.70", "0.80", "0.90", "0.98", "1.00"
]
# all_dratios = ["0.10"]

a0s = [8, 16, 23, 30, 42]

for geom in ['mixed_30nc']:
    #for step in range(1000, 1001, 1000):   # first step to verify density
    #for step in range(8000, 15001, 1000):  # steps around peak intensity
    for step in range(10000, 10001, 1000):  # peak, prompt electrons dominate 2nd half of target
        for sfilter in ['onLaserAxisNarrow']:  # 'all', 'onLaserAxisNarrow', 'onLaserAxisWide', 'onLaserAxis', 'fwPinhole', 'fwRCF']:
            #print(sfilter, step, geom)
            plotAndSave(sfilter, a0s, step, geom, dratios=all_dratios)
