#!/usr/bin/env python
#
# Copyright 2018-2019 Axel Huebl
#
# License: CC-BY-SA 4.0 International
#
# This script is mainly used to estimate the laser accelerated electrons
# and which average kinetic energy <T_e> and density n_e they correspond
# to.
# Unfortunately, there is no well documented workflow (nor theory) in our
# community to measure (predict) both quantities self-consistently and at
# the same time for a wide range of parameters. Therefore, consider this
# a best-effort for a ballpark number and trying to be transparent about
# how the numbers were generated.
#
# Alternatively, one can also broadly plug in the absorption for n_e/n0 and
# measure the <T_e> as below or take it from one of the theoretical models,
# such as Kluge (2011).
#
import matplotlib as mpl
#mpl.use('Agg') # non-interactive backend
import matplotlib.pyplot as plt
from matplotlib.colors import colorConverter, LogNorm, Normalize
from picongpu.plugins.data import PhaseSpaceData
from scipy.special import lambertw
from scipy.optimize import fsolve, curve_fit
import numpy as np
import pandas as pd
import pickle
from collections import defaultdict
from os import sys, path


lon_mid_y = 4.  # microns
L_x = 7488      # number of cells, transversal box

# selection for analysis of temperature and ene dens
low_dy_mid = -0.0   # microns
hig_dy_mid = +0.9   # microns
mom_cutoff = {      # [beta*gamma] above this we consider e- as "hot"/"laser"/"prompt"
                    #   we correct for the cutoff by assuming a Boltzmann distribution
     8: 0.2,  # position of "kink" in py(e-) from cold to laser/prompt e-
  #  8: 2.0,  #
    16: 1.0,  # here and below:
    23: 1.0,  #
    30: 1.0,  #    just select "all relativistic e-"
    42: 1.0,  #
}
mom_high_cutoff = 99.e9
include_return_currents = {  # calc similar to absorption
     8: True,  # True: account for bulk influence: slower expansion, re-heating
  #  8: False,
    16: False,
    23: False,
    30: False,
    42: False,
}
sfilter_spatial = {  # 2D effects
     8: 'onLaserAxisNarrow',
    16: 'onLaserAxisNarrow',
    23: 'onLaserAxis',
    30: 'onLaserAxisWide',
    42: 'onLaserAxisWide',
}

step_peak_target = 8190  # 64.3 fs
dt = 3.33e-9 / ( 1.415 * 2.99792458e8 )  # 7.84995e-18  # s
tau = 30.e-15  # s

lambda_0 = 800.e-9  # m
n0 = 30.            # n_c
mu = 1.e6
nc = 1.11485e21 * 1.e6 / (lambda_0 * mu) / (lambda_0 * mu)

q_e = 1.602e-19      # C
MeV2J = 1.60218e-13  # J

myCmaps = {}
colors = {"H": "blue", "D": "darkorange", "e": "red"}

# see https://stackoverflow.com/a/8702435/2719194
#     https://stackoverflow.com/a/26497091/2719194
def nested_dict(): return defaultdict(nested_dict)

# generate the colors for colormaps
def generateCmaps():
    for s in colors.keys():
        myColor1 = colorConverter.to_rgba('white', alpha=1.0)
        myColor2 = colorConverter.to_rgba(colors[s], alpha=1.0)
        myCmap = mpl.colors.LinearSegmentedColormap.from_list(
            'my_cmap_' + s, [myColor1,myColor2], 1200)
        myCmap._init() # create the _lut array, with rgba values
        myCmaps[s] = myCmap

def getData(species, sfilter, a0, step, geom, dratio, ps_name='ypy'):
    """..."""
    phase_space = PhaseSpaceData("a0_" + str(a0) + "/2D_" + geom + "_dratio" + dratio)
    ps, ps_meta = phase_space.get(species, sfilter, iteration=step, ps=ps_name)

    # convert to momentum per nukleon
    #if species == 'D':
    #    ps_meta.extent = ps_meta.extent * [1., 1., 0.5, 0.5]

    return ps, ps_meta

def getEfield(a0, step, geom, dratio, component='y'):
    """..."""
    import h5py as h5
    sim_dir = "a0_" + str(a0) + "/2D_" + geom + "_dratio" + dratio + "/"
    f = h5.File(sim_dir + "simOutput/h5/fields_" + str(step) + ".h5", "r")

    E_y = f["/data/" + str(step) + "/fields/E/" + component]
    s = E_y.shape

    # single line sliced out
    #          y, x
    # return E_y[:, E_y.shape[1]//2], E_y.attrs["unitSI"]

    # sfilter == 'onLaserAxis' means
    #   0.45 to 0.55 around center axis
    x_min = int(E_y.shape[1] * 0.45)
    x_max = int(E_y.shape[1] * 0.55)
    E_y_slice = E_y[:, x_min:x_max]
    return np.average(E_y_slice, axis=1), E_y.attrs["unitSI"]

def plotAndSave(a0s, step, geom, dratios):
    """..."""
    db_ne0 = {}  # n0
    db_Te  = {}  # MeV
    db_we  = {}  # MeV * n_c
    db_ne0[step] = {}
    db_Te[step] = {}
    db_we[step] = {}
    for a0 in a0s:
        sfilter = sfilter_spatial[a0]
        db_ne0[step][a0] = []
        db_Te[step][a0]  = []
        db_we[step][a0]  = []

        for dratio in dratios:
            f = plt.figure(figsize=[4, 3])
            # left, bottom, W, H
            ax = f.add_axes([0.18, 0.17, 0.40, 0.7])
            ax_py = f.add_axes([0.58, 0.17, 0.10, 0.7])
            ax_Ey = f.add_axes([0.68, 0.17, 0.10, 0.7])
            cax_e = f.add_axes([0.79, 0.17, 0.03, 0.7])
            #cax_H = f.add_axes([0.79, 0.17, 0.03, 0.7])
            # ax = plt.gca()

            e_ps, e_ps_meta = getData('e', sfilter, a0, step, geom, dratio)
            e_ps = np.abs(e_ps) * e_ps_meta.dV / q_e  # normalize to count

            # unit conversion from SI
            mu = 1.e6  # meters to microns
            e_mc_r = 1. / (9.109e-31 * 2.9979e8)  # electrons: kg * m / s to beta * gamma
            #H_mc_r = e_mc_r / 1836.152672         # protons: kg * m / s to beta * gamma
            #D_mc_r = H_mc_r / 1.9985              # protons: kg * m / s to beta * gamma

            # plotting
            ax.axhline(0.0,
                alpha=0.1,
                lw=1,
                color='black')
            ax.axvline(3.0,
                alpha=0.1,
                lw=1,
                color='grey')
            ax.axvline(5.0,
                alpha=0.1,
                lw=1,
                color='grey')

            im_e = ax.imshow(
                e_ps.T,
                extent = e_ps_meta.extent * [mu, mu, e_mc_r, e_mc_r],
                interpolation = 'nearest',
                aspect = 'auto',
                origin='lower',
                norm = LogNorm(),
                cmap=myCmaps['e'],
                label='e',
                #vmin=1.e-28,
                #vmax=1.e-23
            )

            dy = e_ps_meta.dV**(1./3.)
            y_bins = np.linspace(
                e_ps_meta.extent[0] * mu,
                e_ps_meta.extent[1] * mu,
                e_ps.shape[0])

            # second half of the target
            low_y_bin = np.argmax(y_bins > lon_mid_y + low_dy_mid)
            hig_y_bin = np.argmax(y_bins > lon_mid_y + hig_dy_mid)

            py_hist = np.sum(e_ps[low_y_bin:hig_y_bin, :-1], axis=0)

            # divide by sfilter (onLaserAxis) volume
            ps_V_cells = 1.0
            if sfilter == "onLaserAxis":
                #   0.45 to 0.55 around center axis
                ps_V_cells = (
                    L_x * 0.1 *                # Nx
                    (hig_y_bin - low_y_bin) *  # Ny
                    1                          # Nz
                )
            elif sfilter == "onLaserAxisNarrow":
                #   0.495 to 0.505 around center axis
                ps_V_cells = (
                    L_x * 0.01 *               # Nx
                    (hig_y_bin - low_y_bin) *  # Ny
                    1                          # Nz
                )
            elif sfilter == "onLaserAxisWide":
                #   0.38 to 0.62 around center axis
                ps_V_cells = (
                    L_x * 0.24 *               # Nx
                    (hig_y_bin - low_y_bin) *  # Ny
                    1                          # Nz
                )
            elif sfilter == "all":
                ps_V_cells = (
                    L_x *                      # Nx
                    (hig_y_bin - low_y_bin) *  # Ny
                    1                          # Nz
                )
            py_hist = py_hist / ps_V_cells

            py_bins = np.linspace(
                e_ps_meta.extent[2] * e_mc_r,
                e_ps_meta.extent[3] * e_mc_r,
                py_hist.shape[0])

            eney_bins = 0.511 * ((1.+py_bins**2)**0.5 - 1.)  # MeV
            eney_bins[py_bins < 0.]  *= -1.0
            deney = eney_bins[1:] - eney_bins[:-1]
            eney_bins = eney_bins[:-1]

            # mean for:
            # -[x] gamma_e > 0.5**0.5
            # -[x] lon_mid_y + 0.9 > y > lon_mid_y - 0.9

            # momentum [beta*gamma]
            # beta * gamma = sqrt(gamma**2 - 1)
            ene_cutoff = 0.511 * ((1.+mom_cutoff[a0]**2)**0.5 - 1.)  # MeV
            ene_high_cutoff = 0.511 * ((1.+mom_high_cutoff**2)**0.5 - 1.)  # MeV

            low_py_bin = np.argmax(py_bins > mom_cutoff[a0])       # also the low_eney_bin
            hig_py_bin = np.argmax(py_bins > mom_high_cutoff)  # also the hig_eney_bin
            if hig_py_bin == 0: # select all
                hig_py_bin = None

            #low_py_bin = 0  # debug: should now be full density at t = 0
            #hig_py_bin = -1

            # include return currents:
            if include_return_currents[a0]:
                zero_y_bin = np.argmax(py_bins >= 0)
                neg_bins = py_hist[:zero_y_bin].copy()
                py_hist[zero_y_bin:zero_y_bin+neg_bins.shape[0]] += neg_bins[::-1]
                py_hist[:zero_y_bin] = 0.0

            # histogram p -> Ene
            eney_hist = py_hist[:-1] / deney  # normalize bin size: count / MeV
            ax_Ey.plot(
                eney_hist,
                eney_bins,
                color=colors['e']
            )

            # calc start
            mom_mean_up = np.sum(py_hist[low_py_bin:hig_py_bin] * py_bins[low_py_bin:hig_py_bin]) / np.sum(py_hist[low_py_bin:])

            # energy [MeV]
            # careful: sqrt(1+<bg>^2) != <sqrt(1+bg^2)>
            ene_mean_up = np.sum(eney_hist[low_py_bin:hig_py_bin] * eney_bins[low_py_bin:hig_py_bin]) / np.sum(eney_hist[low_py_bin:hig_py_bin])
            ne_mean_up = np.sum(py_hist[low_py_bin:hig_py_bin]) / e_ps_meta.dV
            # wV: energy density * dV, Ne: count
            wV_int_up = np.trapz(eney_hist[low_py_bin:hig_py_bin] * eney_bins[low_py_bin:hig_py_bin], eney_bins[low_py_bin:hig_py_bin])
            Ne_int_up = np.trapz(eney_hist[low_py_bin:hig_py_bin], eney_bins[low_py_bin:hig_py_bin])
            # correction for cutoff (mom_cutoff) for f_e(E) = Ne/kB/T * exp(-E/kB/T)

            #print("gamma_cutoff {:.2f}, {:.2f} MeV".format( (1.+mom_cutoff**2)**0.5, ene_cutoff) )

            #ne_mean_up = Ne_int_up / ene_mean_up
            #print(Ne_int_up / e_ps_meta.dV / nc / n0)

            #  dV<w> = int E * f(E) dE
            ene_mean = wV_int_up / Ne_int_up - ene_cutoff
            # N_e,0 = int f(E) dE
            ne_mean = Ne_int_up * np.exp(ene_cutoff / ene_mean) / e_ps_meta.dV

            # <x> = int f(x) dx
            #lw = lambertw(ene_cutoff/ene_mean_up)
            #if lw.imag != 0.0:
            #    print("WARNING: lambertw correction is not purely real!")
            #ene_mean = ene_cutoff / lw.real
            #print(ene_cutoff / lw.real)

            # <x> = int x * f(x) dx
            #f_e = lambda ene_mean: (
            #    ene_mean - (ene_cutoff + ene_mean) * np.exp(-ene_cutoff/ene_mean)
            #    + ene_mean_up / ene_mean
            #    - 1.
            #)
            #initial_guess = ene_int_up
            #ene_mean = fsolve(f_e, initial_guess)
            #if len(ene_mean) != 1:
            #    print("WARNING: more than one root of f_e!")
            #ene_mean = ene_mean[0]

            #print("up {:.2f} -> lam {:.2f} -> new {:.2f}".format(
            #    ene_mean_up, ene_cutoff / lw.real, ene_mean))

            # density
            #ne_y = np.sum(e_ps[low_y_bin:hig_y_bin, low_py_bin:hig_py_bin], axis=1)

            #ne_mean = np.mean(ne_y[:]) / ps_V_cells

            #print(ne_mean / nc / n0)
            # correction for cutoff (mom_cutoff) for n_e(E) = n_e,0 * exp(-E/kB/T)
            # n<E> = n* int E * f(E) dE
            #ne_mean = nc * ene_mean * (1. - np.exp(-ene_cutoff/ene_mean)) + ne_mean
            #ne_mean += nc * n0 * ene_mean * (1. - np.exp(-ene_cutoff/ene_mean))
            # ne_mean * ene_mean_up / ene_mean / (1. - (ene_mean - (ene_cutoff+ene_mean)*np.exp(-ene_cutoff/ene_mean)))
            #ne_mean = ne_mean / ene_mean * np.exp(ene_cutoff / ene_mean)

            #ene_dens_up = np.sum(py_hist[low_py_bin:hig_py_bin] / ps_V_cells / UNIT_LENGTH * py_hist[low_py_bin:hig_py_bin] * ene_bins) / np.sum(py_hist[low_py_bin:hig_py_bin])

            #ene_dens_up = ene_dens_up
            #ne_mean = ene_dens_up / ene_mean * np.exp(ene_cutoff / ene_mean)

            # Energy Density [MeV n_c]
            # w_e = ene_mean_up * ne_mean_up / nc  # cut distribution
            w_e = ene_mean * ne_mean / nc          # Boltzmann distribution (BD)

            print("a0={:02d} t={:05d} d={:.2f}: {:.2f} MeV ({:.2f} BD) @ n_e = {:.3f} n_0 ({:.3f} BD), w_e = {:.2f} MeV n_c".format(
                a0,
                step,
                float(dratio),
                ene_mean_up,
                ene_mean,
                ne_mean_up / nc / n0,
                ne_mean / nc / n0,
                w_e
            ))

            db_ne0[step][a0].append(ne_mean / nc / n0)  # n0
            db_Te[step][a0].append(ene_mean)            # MeV
            db_we[step][a0].append(w_e)                 # MeV * n_c

            ax_py.plot(
                py_hist,
                py_bins,
                color=colors['e']
            )

            ax_py.axhline(
                y = mom_mean_up
            )
            ax_py.axhline(
                y = mom_cutoff[a0],
                color='grey',
                ls=':'
            )
            ax_py.axhline(
                y = mom_high_cutoff,
                color='black',
                ls=':'
            )
            ax_Ey.axhline(
                y = ene_mean
            )
            ax_Ey.axhline(
                y = ene_cutoff / 0.511,
                color='grey',
                ls=':'
            )
            ax_Ey.axhline(
                y = ene_high_cutoff / 0.511,
                color='black',
                ls=':'
            )

            ax_py.set_yticklabels([])
            ax_py.set_xscale("log")
            ene_org_lim = ax_py.get_xlim()
            ax_py.set_xlim([ene_org_lim[1], ene_org_lim[0]])
            ax_py.set_ylim([py_bins[0], py_bins[-1]])
            ax_py.set_xlabel(r"$p_y$")

            ax_Ey.set_yticklabels([])
            ax_Ey.set_xscale("log")
            ene_org_lim = ax_Ey.get_xlim()
            ax_Ey.set_xlim([ene_org_lim[1], ene_org_lim[0]])
            ax_Ey.set_ylim([eney_bins[0], eney_bins[-1]])
            ax_Ey.set_xlabel(r"$K_y$")

            cbar_e = plt.colorbar(im_e, cax=cax_e)
            #cbar_D = plt.colorbar(im_D, cax=cax_D)
            #cbar_H.ax.set_yticklabels([])
            # cbar_H.set_label(r'$Q / \mathrm{d}r \mathrm{d}p$ [$\mathrm{C s kg^{-1} m^{-2}}$]')
            cbar_e.set_label(r'$1 / \mathrm{d}r \mathrm{d}p$ [$\mathrm{s kg^{-1} m^{-2}}$]')

            # limits
            if step <= 15000:
                ax.set_xlim([2.8, 7.5])
            #elif step <= 20000:
            #    ax.set_xlim([2.8, 9.5])
            #elif step <= 25000:
            #    ax.set_xlim([2.8, 11.])

            #if step <= 34000:
            #    ax.set_xlim([2.4, 11.])
            #elif step <= 50000:
            #    ax.set_xlim([0., 19.])

            # fw filter only? py >= 0
            if sfilter.startswith("fw"):
                p_ax_range = list(ax.get_ylim())
                p_ax_range[0] = 0.0
                p_ax_range[1] *= 0.6
                ax.set_ylim(p_ax_range)

            # annotations
            ax.set_xlabel(r'${0}$'.format(e_ps_meta.r) + r' [$\mathrm{\mu m}$]')
            # FIXME: is this scaling useful for multi-ion compares?!
            ax.set_ylabel(r'$p_{0}$ [$\beta\gamma$]'.format(e_ps_meta.p))

            #f.text(s='e',
            #    color=colors['e'],
            #    x=0.68, y=0.25)
            #f.text(s='H',
            #    color=colors['H'],
            #    x=0.68, y=0.20)

            #ax.title("e & {0} H ({1}) @ t = {2}".format(dratio, sfilter, step))
            #ax.legend()

            #plt.show()
            fname = (geom + "/ps_ypy/a0_" + str(a0) + "/a0_" +
                str(a0) + "_e_" + sfilter + "_" +
                geom +
                "_dratio" + dratio +
                "_ypy_")
            f.savefig(fname + str(step) + ".svg", dpi=300)
            plt.close(f)

        # stdout block for each a0
        print("")

    # print summary
    print("Summary:")
    ne0_avg = {}
    ne0_std = {}
    Te_avg  = {}
    Te_std  = {}
    we_avg  = {}
    we_std  = {}
    for s, v in db_ne0.items():
        print("t={:05d}".format(s))
        for a0 in a0s:
            print("  a0={:02d}".format(a0))
            # removes d=0.98 for equally binned statistics
            r098 = dratios.index("0.98") if "0.98" in dratios else np.intp([])

            ne0_avg[a0] = np.mean(np.delete(db_ne0[s][a0], r098))
            ne0_std[a0] = np.var(np.delete(db_ne0[s][a0], r098))**0.5
            Te_avg[a0]  = np.mean(np.delete(db_Te[s][a0], r098))
            Te_std[a0]  = np.var(np.delete(db_Te[s][a0], r098))**0.5
            we_avg[a0]  = np.mean(np.delete(db_we[s][a0], r098))
            we_std[a0]  = np.var(np.delete(db_we[s][a0], r098))**0.5
            print("    Te = {:.2f} +/- {:0.2f} MeV, "
                  "ne = {:.3f} +/- {:0.3f} n0, "
                  "we = {:.2f} +/- {:0.2f} MeV nc".format(
                Te_avg[a0],  Te_std[a0],
                ne0_avg[a0], ne0_std[a0],
                we_avg[a0],  we_std[a0]
            ))
    print("")

    # pickle results
    save_name = geom + '/ps_ypy/e_Te' # + '_' + sfilter
    saved_results = nested_dict()
    try:
        # restore other geometries if already there
        with open(save_name + '.p', 'rb') as fp:
            saved_results = pickle.load(fp)
    except (IOError, EOFError):
        pass

    for s, v in db_ne0.items():
        for a0 in a0s:
            saved_results[s][a0]["d"] = np.array(dratios).astype(float)
            saved_results[s][a0]["ne0"] = np.array(db_ne0[s][a0])
            saved_results[s][a0]["Te"] = np.array(db_Te[s][a0])
            saved_results[s][a0]["we"] = np.array(db_we[s][a0])
            saved_results[s][a0]["ne0_avg"] = ne0_avg
            saved_results[s][a0]["Te_avg"] = Te_avg
            saved_results[s][a0]["we_avg"] = we_avg
            saved_results[s][a0]["ne0_std"] = ne0_std
            saved_results[s][a0]["Te_std"] = Te_std
            saved_results[s][a0]["we_std"] = we_std

    with open(save_name + '.p', 'wb') as fp:
        pickle.dump(saved_results, fp)

    # plot summary
    #db_ne0[step][a0]  # n0
    #db_Te0[step][a0]  # MeV
    #db_we[step][a0]   # MeV * n_c

generateCmaps()

all_dratios = [
    "0.00", "0.10", "0.20", "0.30", "0.40", "0.50",
    "0.60", "0.70", "0.80", "0.90", "0.98", "1.00"
]
# all_dratios = ["0.10"]

a0s = [8, 16, 23, 30, 42]

for geom in ['mixed_30nc']:
    #for step in range(1000, 1001, 1000):   # first step to verify density
    #for step in range(8000, 15001, 1000):  # steps around peak intensity
    for step in range(10000, 10001, 1000):  # peak, prompt electrons dominate 2nd half of target
        plotAndSave(a0s, step, geom, dratios=all_dratios)
