#!/usr/bin/env python
#
# Copyright 2018-2019 Axel Huebl
#
# License: CC-BY-SA 4.0 International
#
import matplotlib as mpl
#mpl.use('Agg') # non-interactive backend
import matplotlib.pyplot as plt
import matplotlib.ticker as mtick
from matplotlib.patches import ConnectionPatch
from matplotlib.colors import colorConverter, LogNorm, Normalize
import h5py as h5
import numpy as np


# a0=16  # 16, 30, 42

# converged ion energy, before particles leave the box
step={
    16: 75000,
    30: 60000,
    42: 45000
}
# laser peak on target: step 8190 (64fs)
# max energy in e-: ~100.e-15 / 7.84995e-18 == 12739
# step 10'000 =  14fs after peak @8190
# step 15'000 =  53fs after peak @8190
# step 20'000 =  93fs after peak @8190
# step 25'000 = 132fs after peak @8190
step_e = 10000

EneLimMeV={  # ions
     8: 20.,
    16: 40.,
    23: 60.,
    30: 80.,
    42: 160.
}
#EneLimMeV={  # electrons
#    16:  8.,
#    30: 20.,
#    42: 35.
#}

# sfilter='all'      # 'all', 'originFront', 'originRear'
geom='mixed_30nc'  # mixed_30nc separ_30nc  separ_300nc separ_1000nc separ_1700nc

oneMevLineouts = {  # ions
     8: [5.],
    16: [9.],  # 18.
    23: [18.],
    30: [22.], # 42.
    42: [34.], # 70.
}
#oneMevLineouts = {  # electrons
#    16: [2.],
#    30: [6.],
#    42: [10.],
#}
dMeVLineouts = 1.0
lsMeVLineouts = ['-', '--']  # ':', '-.'

pointingLineouts = {
    16: [0.,  -7.],
    30: [0., -11.],
    42: [0., -15.],
}
dDegreeLineouts = 4.0  # +/- 2 degree
lsDegreeLineouts = ['-.', '--']  # '-', ':', '-.'
cDegreeLineouts = ['black', 'grey']

all_dratios=["0.00", "0.10", "0.50", "0.60", "0.70", "0.80", "0.90", "0.98", "1.00"]

myCmaps = {}
colors = {"H": "blue", "D": "darkorange", "e": "red", "front": "green"}

# generate the colors for colormaps
def generateCmaps():
    for s in colors.keys():
        myColor1 = colorConverter.to_rgba('white', alpha=0.5)
        myColor2 = colorConverter.to_rgba(colors[s], alpha=1.0)
        myCmap = mpl.colors.LinearSegmentedColormap.from_list(
            'my_cmap_' + s, [myColor1,myColor2], 1200)
        myCmap._init() # create the _lut array, with rgba values
        myCmaps[s] = myCmap

def getData(species, sfilter, a0, step, geom, dratio, suffix=""):
    """..."""
    h5f = h5.File(
        "a0_" + str(a0) +
        "/2D_" + geom + "_dratio" + dratio + suffix +
        "/simOutput/" + species + "_calorimeter/" +
        sfilter + "/" +
        species + "_calorimeter_" +
        sfilter + "_" + sfilter + "_" +
        str(step) + "_0_0_0.h5",
        "r")

    data = h5f["/data/{0}/calorimeter".format(str(step))]

    # [energy, pitch (zy), yaw (xy)]
    counts = data[()]

    # attributes
    ene_range = np.array([data.attrs["minEnergy[keV]"], data.attrs["maxEnergy[keV]"]])
    max_pitch = data.attrs["maxPitch[deg]"]
    max_yaw = data.attrs["maxYaw[deg]"]
    posPitch = data.attrs["posPitch[deg]"]
    posYaw = data.attrs["posYaw[deg]"]
    unit_energy = 7.89934e-12
    unitSI = data.attrs["unitSI"] / unit_energy
    logScale = data.attrs["logScale"]

    h5f.close()

    if species == 'D':
        ene_range = ene_range * 0.5

    return counts * unitSI, ene_range, max_pitch, max_yaw

def get1MeVslices(lineouts_MeV, yaw_ene, ene_range_keV):
    """..."""
    v = {}
    ene_bins_MeV = np.linspace(
        ene_range_keV[0] * 1.e-3,
        ene_range_keV[1] * 1.e-3,
        yaw_ene.shape[0])
    for oneMeVlow in lineouts_MeV:
        low = np.argmax(ene_bins_MeV >= oneMeVlow)
        up = np.argmax(ene_bins_MeV > oneMeVlow + dMeVLineouts)
        v[oneMeVlow] = np.sum(yaw_ene[low:up, :], axis=0)
    return v

def getPinholeSlices(pointingLineouts_deg, yaw_ene, max_yaw_deg):
    """..."""
    v = {}
    yaw_bins_deg = np.linspace(
        max_yaw_deg * -1.,
        max_yaw_deg,
        yaw_ene.shape[1])
    for oneDegLog in pointingLineouts_deg:
        low = np.argmax(yaw_bins_deg >= oneDegLog - dDegreeLineouts/2.)
        up = np.argmax(yaw_bins_deg > oneDegLog + dDegreeLineouts/2.)
        v[oneDegLog] = np.sum(yaw_ene[:, low:up], axis=1)
    return v

def yawCropSlice(species, species_list, shape, inverse=False):
    """select upper or (inverse) lower half of a species"""
    if len(species_list) == 2:
        if (species == species_list[0]) ^ inverse:
            return np.s_[shape[1]//2:]
        else:
            return np.s_[:shape[1]//2]
    elif inverse:
        assert("Not possible!")
    else:
        return np.s_[:]

def plotAndSave(species_list, sfilter_list, a0, step, geom, dratios, suffix=""):
    """..."""
    if type(sfilter_list) is not list:
        sfilter_list = [sfilter_list]
    if type(species_list) is not list:
        species_list = [species_list]

    for dratio in dratios:
        f = plt.figure(figsize=[4, 3])

        # left, bottom, W, H
        # ax = f.add_axes([0.15, 0.15, 0.65, 0.55])
        ax = f.add_axes([0.15, 0.15, 0.50, 0.55])
        cax = f.add_axes([0.83, 0.15, 0.03, 0.55])
        ax_yaw = f.add_axes([0.15, 0.75, 0.65, 0.15])
        ax_pnt = f.add_axes([0.68, 0.15, 0.12, 0.55], sharey=ax)

        lineout = {}
        lineoutPtn = {}
        for species in species_list:
            for sfilter in sfilter_list:
                counts, ene_range, max_pitch, max_yaw = \
                    getData(species, sfilter, a0, step, geom, dratio, suffix)
                yaw_ene = counts[:, 0, :]

                # sum filters, slice for selected 1 MeV ranges
                if species not in lineout:
                    lineout[species] = {}
                oneMevSlices = get1MeVslices(oneMevLineouts[a0], yaw_ene, ene_range)
                for oneMeVlow in oneMevLineouts[a0]:
                    if oneMeVlow not in lineout[species]:
                        lineout[species][oneMeVlow] = 0.0
                    lineout[species][oneMeVlow] += oneMevSlices[oneMeVlow]

                # sum filters, slice for +/-2 degree ranges
                if species not in lineoutPtn:
                    lineoutPtn[species] = {}
                ptnSlices = getPinholeSlices(pointingLineouts[a0], yaw_ene, max_yaw)
                for onePtnLow in pointingLineouts[a0]:
                    if onePtnLow not in lineoutPtn[species]:
                        lineoutPtn[species][onePtnLow] = 0.0
                    lineoutPtn[species][onePtnLow] += ptnSlices[onePtnLow]

                # side-by-side comparisons: crop away half of the opening in yaw
                if len(species_list) == 2:
                    yaw_ene[:, yawCropSlice(species, species_list, yaw_ene.shape, True)] = 0.0

                sfilter_cmap = myCmaps[species]
                if len(sfilter_list) > 1 and sfilter == 'originFront':
                    sfilter_cmap = myCmaps['front']

                if species_list[0] == 'e':
                    vmin = 1.e-2
                    vmax = 1.e3
                else:
                    vmin = 1.e-1
                    vmax = 1.e6
                im = ax.imshow(
                    np.abs(yaw_ene),
                    extent = [-max_yaw, max_yaw,
                              ene_range[0] * 1.e-3, ene_range[1] * 1.e-3],
                    interpolation = 'nearest',
                    aspect = 'auto',
                    origin='lower',
                    norm = LogNorm(),
                    zorder=3,
                    cmap=sfilter_cmap,
                    vmin=vmin,
                    vmax=vmax
                )

        # single RCF stack layer
        i = 0
        for oneMeVlow in oneMevLineouts[a0]:
            ax.axhline(y=oneMeVlow + dMeVLineouts/2.,
                zorder=1,
                linewidth=1.0,
                linestyle=lsMeVLineouts[i],
                color='grey')
            ax.text(s="{0}-{1} MeV".format(int(oneMeVlow), int(oneMeVlow+dMeVLineouts)),
                color='grey',
                size=ax.yaxis.label.get_fontsize() * 0.75,
                x=-85., y=oneMeVlow + dMeVLineouts/2. + EneLimMeV[a0] * 0.04)
            for species in species_list:
                # underplot with grey
                for c in ['grey', colors[species]]:
                    ls = lsMeVLineouts[i]
                    if c == 'grey' : ls = '-'
                    lb = "{0}-{1}MeV".format(oneMeVlow, oneMeVlow + dMeVLineouts)
                    if c != 'grey' : lb = None
                    # sel = yawCropSlice(species, species_list, yaw_ene.shape)
                    sel = np.s_[:]
                    ax_yaw.plot(
                        np.linspace(-max_yaw, max_yaw, lineout[species][oneMeVlow].shape[0])[sel],
                        lineout[species][oneMeVlow][sel] * 1.e-4,  # just too make axes easier
                        linestyle=ls,
                        zorder=3,
                        label=lb,
                        color=c
                    )
            i += 1

        # pinhole lineout
        i = 0
        for pointingLine in pointingLineouts[a0]:
            ax.axvline(x=pointingLine,
                zorder=1,
                linewidth=1.0,
                linestyle=lsDegreeLineouts[i],
                color=cDegreeLineouts[i])
            for oDeg in [dDegreeLineouts/2., -dDegreeLineouts/2.]:
                x = pointingLine + oDeg
                ax_yaw.axvline(x=x,
                    zorder=1,
                    linewidth=1.0,
                    linestyle=lsDegreeLineouts[i],
                    color=cDegreeLineouts[i])
                con = ConnectionPatch(
                    xyA=(pointingLine, EneLimMeV[a0]), xyB=(x, 0.),
                    coordsA="data", coordsB="data",
                    zorder=5,
                    axesA=ax, axesB=ax_yaw, color=cDegreeLineouts[i],
                    linestyle=lsDegreeLineouts[i], linewidth=1.0)
                ax.add_artist(con)

            if 'H' in species_list:
                lntSpec = 'H'
            else:
                lntSpec = species_list[0]

            lb = r"{0}$\degree$".format(int(abs(pointingLine)))
            ene_bins = np.linspace(ene_range[0] * 1.e-3, ene_range[1] * 1.e-3,
                    lineoutPtn[lntSpec][pointingLine].shape[0])
            ax_pnt.plot(
                lineoutPtn[lntSpec][pointingLine],
                ene_bins,
                linestyle=lsMeVLineouts[i],
                zorder=3,
                label=lb,
                color=cDegreeLineouts[i]
            )
            print("Mean {0} energy at {1} degrees: {2:2f} MeV".format(
                lntSpec,
                int(abs(pointingLine)),
                np.sum(lineoutPtn[lntSpec][pointingLine] * ene_bins) /
                np.sum(lineoutPtn[lntSpec][pointingLine])
            ))

            i += 1

        cbar = plt.colorbar(im, cax=cax)
        cbar.solids.set_edgecolor("face")
        cbar.ax.set_title(
            'count\n/ MeV\n/ solid angle\n[arb. u.]',
            loc='left',
            size=ax.yaxis.label.get_fontsize() * 0.75)
        # cbar.ax.set_label_coords(0.92, 0.8)

        # fwRCF range
        for axl in [ax, ax_yaw]:
            axl.axvline(x=-25.,
                linestyle=':',
                linewidth=1.0,
                zorder=4,
                color='black')
            axl.axvline(x=25.,
                linestyle=':',
                linewidth=1.0,
                zorder=4,
                color='black')
        ax_yaw.axvline(x=-2.,
            linestyle='-.',
            linewidth=1.0,
            zorder=4,
            color='black')
        ax_yaw.axvline(x=2.,
            linestyle='-.',
            linewidth=1.0,
            zorder=4,
            color='black')
        ax.text(s='RCF selection', color='black',
                size=ax.yaxis.label.get_fontsize() * 0.75,
                rotation=90,
                x=3., y=EneLimMeV[a0] * 0.90)
        ax_yaw.text(s='pinhole', color='black',
                size=ax_yaw.yaxis.label.get_fontsize() * 0.65,
                rotation=90,
                x=-1.0, y=ax_yaw.get_ylim()[1] * 0.7)

        # species crop
        if len(species_list) > 1:
            ax.axvline(x=0.,
                linestyle='-.',
                linewidth=1.0,
                zorder=4,
                color='black')
            ax.text(s=species_list[0] + '$\mathregular{{}_{rear}}$',
                color=colors[species_list[0]],
                x=18., y=EneLimMeV[a0]* 0.8)
            ax.text(s=species_list[0] + '$\mathregular{{}_{front}}$',
                color=colors['front'],
                x=18., y=EneLimMeV[a0]* 0.65)
            ax.text(s=species_list[1] + '$\mathregular{{}_{rear}}$',
                color=colors[species_list[1]],
                x=-80., y=EneLimMeV[a0] * 0.8)
            ax.text(s=species_list[1] + '$\mathregular{{}_{front}}$',
                color=colors['front'],
                x=-80., y=EneLimMeV[a0] * 0.65)

        ax.text(s="$\mathregular{{a}_0}$ = " + "{0}".format(a0),
                color='red',
                ha='left',
                x=7., y=EneLimMeV[a0] * 0.31)

        # limits
        ax.set_ylim([0., EneLimMeV[a0]])
        ax.set_xlim([-max_yaw, max_yaw * 0.50 / 1.44 / 0.65])
        ax_yaw.set_xlim([-30., 30.])  # max_yaw
        ax_yaw.set_ylim([0., None])
        if species_list[0] == 'e':
            ax_pnt.set_xlim([1.e1, 1.e5])
        else:
            ax_pnt.set_xlim([1.e1, 1.e7])
        ax_pnt.set_xscale("log")
        ene_org_lim = ax_pnt.get_xlim()
        ax_pnt.set_xlim([ene_org_lim[1], ene_org_lim[0]])

        # annotations
        species_name = "+".join(species_list)
        sfilter_name = "+".join(sfilter_list)

        for x in [-25., 25.]:
            con = ConnectionPatch(
                xyA=(x, EneLimMeV[a0]), xyB=(x, 0.),
                coordsA="data", coordsB="data",
                zorder=5,
                axesA=ax, axesB=ax_yaw, color="black",
                linestyle=':', linewidth=1.0)
            ax.add_artist(con)

        leg = ax_pnt.legend(handlelength=0, handletextpad=0, borderpad=0, borderaxespad=0.35,
            loc='upper left', framealpha = 0)  # , fancybox=True
        #renderer = f.canvas.get_renderer()
        #shift = max([t.get_window_extent(renderer).width for t in leg.get_texts()])
        i = 0
        for text in leg.get_texts():
            plt.setp(text, color = cDegreeLineouts[i])
            text.set_color(cDegreeLineouts[i])
            text.set_ha('left')
            #text.set_position((shift, 0))
            i += 1

        ax.set_xlabel(r'yaw  [deg]')
        ax.set_ylabel(r'Ene / A   [MeV]')
        ax.set_xticks([-90., -60., -30., 0., 30.])
        ax_pnt.set_xlabel(r'H count  [arb. u.]',
            size=ax.yaxis.label.get_fontsize() * 0.85)
        ax_pnt.tick_params(labelleft=False)
        ax_yaw.xaxis.tick_top()
        #                 'count / solid angle\n[arb. u.]'
        if len(oneMevLineouts[a0]) > 1:
            ax_yaw.set_ylabel('count\n[arb. u.]',
                size=ax.yaxis.label.get_fontsize() * 0.85)
        else:
            ax_yaw.set_ylabel('{0}-{1} MeV\ncount [arb. u.]'.format(
                    int(oneMevLineouts[a0][0]), int(oneMevLineouts[a0][0]+dMeVLineouts)
                ), size=ax.yaxis.label.get_fontsize() * 0.80)
        # ax_yaw.yaxis.set_label_coords(0.92, 0.8)

        #plt.title("{0} ({1}) @ t = {2}".format(species_name, sfilter_name, step))
        # ax_yaw.legend()

        #plt.show()
        plt.savefig(geom + "/emissionDistrib/" +
            "a0_" + str(a0) + "_" + species_name + "_" + sfilter_name + "_" +
            geom +
            "_dratio" + dratio +
            "_emissionDistrib.svg", dpi=300)
        plt.close()

generateCmaps()

#for species in ['H', 'D']:
#    plotAndSave(species, sfilter, a0, step[a0], geom, dratios=["0.00", "0.10", "0.50", "0.90", "0.98", "1.00"])

for geom in ['mixed_30nc']:  # 'separ_30nc'
    for a0 in [16, 30, 42]:  # 8, 23
        for sfilter in ['all', ['originFront', 'originRear']]:  # , 'all', 'originFront', 'originRear']:
            # side-by-side species
            plotAndSave(['D', 'H'], sfilter, a0, step[a0], geom, dratios=["0.10", "0.80", "0.90"])

            # individual species
            #plotAndSave('e', sfilter, a0, step_e, geom, dratios=all_dratios)
            #for species in ['H', 'D']:
            #    plotAndSave(species, sfilter, a0, step[a0], geom, dratios=["0.00", "0.10", "0.50", "0.60", "0.70", "0.80", "0.90", "0.98", "1.00"])
