#!/usr/bin/env python
#
# Copyright 2018-2019 Axel Huebl
#
# License: CC-BY-SA 4.0 International
#
import matplotlib as mpl
#mpl.use('Agg') # non-interactive backend
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from os import sys, path


geom='mixed_30nc'      # mixed_30nc separ_30nc separ_300nc separ_1000nc separ_1700nc

species=['e', 'H', 'D']

dt = 7.84995e-18  # s
nC = 1.74e27   # m^-3
n0 = nC * 30.  # m^-3
mec2 = 0.511   # MeV
MeV2J = 1.60218e-13  # J

step_peak_target = 8190
#step_measure = 15000     #  53fs after peak
step_measure = 20000     #  93fs after peak
#step_measure = 25000     # 132fs after peak

# MeV
T_hot_Kluge = lambda a0: ( np.pi*a0/(2.*np.log(4.*a0)) - 1. ) * 0.511  # a0 >> 1 (Kluge, 2011)

def getFieldData(a0, geom, dratio):
    """..."""
    sim_dir = "a0_" + str(a0) + "/2D_" + geom + "_dratio" + dratio

    # step total[Joule] Bx[Joule] By[Joule] Bz[Joule] Ex[Joule] Ey[Joule] Ez[Joule] 
    data = pd.read_csv(
        sim_dir + "/simOutput/fields_energy.dat",
        names = ["step", "total", "Bx", "By", "Bz", "Ex", "Ey", "Ez"],
        dtype = {"step": np.int64, "total": np.float64,
            "Bx": np.float64,
            "By": np.float64,
            "Bz": np.float64,
            "Ex": np.float64,
            "Ey": np.float64,
            "Ez": np.float64
        },
        index_col = 0,
        skiprows = 1,
        sep = " "
    )
    return data

def getData(species, sfilter, a0, geom, dratio):
    """..."""
    sim_dir = "a0_" + str(a0) + "/2D_" + geom + "_dratio" + dratio

    # step Ekin_Joule E_Joule
    data = pd.read_csv(
        sim_dir + "/simOutput/{}_energy_{}.dat".format(species, sfilter),
        names = ["step", "total", "E_plus_m0"],
        dtype = {"step": np.int64, "Ekin": np.float64, "E": np.float64},
        index_col = 0,
        skiprows = 1,
        sep = " "
    )
    return data

def plotAndSave(sfilter, geom, dratios):
    """..."""
    f = plt.figure(figsize=[4, 3])
    # left, bottom, W, H
    ax = f.add_axes([0.15, 0.17, 0.80, 0.7])
    #ax = plt.gca()

    eta = {}
    ene_sum = {}
    for a0 in [8, 16, 23, 30, 42]:
        print("a0 = {}".format(a0))
        eta[a0] = {}
        ene_sum[a0] = {}
        for dratio in dratios:  # ["0.50"]
            data = {}
            d = float(dratio)
            for s in species:
                data[s] = getData(s, sfilter, a0, geom, dratio)

            data["fields"] = getFieldData(a0, geom, dratio)

            # after laser heating phase
            if step_measure > 0:
                for key, val in data.items():
                    data[key] = val[val.index >= step_measure]

            sum_ene = np.zeros(len(data["fields"].index))
            for key, val in data.items():
                vals = val["total"]
                sum_ene += vals
            #    ax.plot(
            #        val.index.values * dt * 1.e12, vals,
            #        label=key)
            #ax.plot(
            #    data["fields"].index.values * dt * 1.e12,
            #    sum_ene,
            #    label="sum")

            #ax.set_xlabel("t  [ps]")
            #ax.legend()
            #plt.show()

            # all energy comes from the laser pulse.
            # Bad estimate: Laser Energy reflects out before full laser is initialized
            #ene_inp = np.max(sum_ene)

            # analytic: energy in 2D laser pulse
            dz = 3.33e-9   # [m]
            tau = 30.e-15  # FWHM Intensity [s]
            w0 = 3.0e-6    # FWHM Intensity [m]
            lambda0 = 800.e-9  # [m]
            I0 = (a0 / 8.549297e-6 / lambda0)**2  # [W / m^2]
            P0 = np.pi * w0**2 / 2. * I0          # 3D [W]
            P0 = w0 * dz / 2. * I0                # 2D [W]
            f_s = 0.94                            # numerical pulse shape factor (Gaussian) [unitless]
            ene_inp = P0 * tau / f_s              # [J]

            # predominantly electrons, but let's be pedantic
            ene_out = np.max(data["e"]["total"] + data["H"]["total"] + data["D"]["total"])
            eta[a0][d] = ene_out / ene_inp
            ene_sum[a0][d] = ene_inp
            print("  d={}: {}".format(dratio, eta[a0][d]))

        d_lst = np.sort([float(d) for d in dratios])
        d_eta = [eta[a0][d] for d in d_lst]
        # laser absorption eta
        ax.plot(
            d_lst,
            d_eta,
            label=r"$a_0$ = {}".format(a0)
        )
        # crude estimate for n^rear_e:
        # eta * a0^2/2 = n_e [nc] * T_e [me c^2]  Kluge 2018, eq. (1) + Kluge 2011
        #ax.plot(
        #    d_lst,
        #    np.array(d_eta) * a0**2 / 2 / (T_hot_Kluge(a0) / mec2),
        #    label=r"$a_0$ = {}".format(a0)
        #)

    #ax.set_xlim([0., 1.])
    ax.set_ylim([0., 1.0])
    ax.set_xlabel(r"$d$")
    ax.set_ylabel(r"$\eta$")
    #ax.set_ylabel(r"$n_e$  [$n_c$]")

    ax.legend(loc="lower center", ncol=3,
        prop={'size': ax.yaxis.label.get_fontsize() * 0.9})

    #plt.show()
    f.savefig(
        geom + "/" + "energy_conversion.svg",
        dpi=300)


all_dratios = [
    "0.00", "0.10", "0.20", "0.30", "0.40", "0.50",
    "0.60", "0.70", "0.80", "0.90", "0.98", "1.00"
]
#all_dratios = ["0.80"]

# step 30'000 is 260fs after peak on target
for geom in ['mixed_30nc']:  # 'separ_30nc'
    for sfilter in ['all']:  # 'all', 'fwPinhole', 'fwRCF'
        print(sfilter, geom)
        plotAndSave(sfilter, geom, dratios=all_dratios)
