#!/usr/bin/env python
#
# Copyright 2018-2019 Axel Huebl
#
# License: CC-BY-SA 4.0 International
#
import matplotlib as mpl
mpl.use('Agg') # non-interactive backend
import matplotlib.pyplot as plt
from matplotlib.colors import colorConverter, LogNorm, Normalize
import h5py as h5
import numpy as np
import json


# ...k: ...fs after peak on target
step={
    16: 75000,
    30: 60000,
    42: 45000
}
# sfilter='fwRCF'      # 'fwRCF', 'fwPinhole', 'all'

# max energy [MeV]
max_energy={
    16: 40.,
    30: 80.,
    42: 160.
}

Nene = 150
dx = 3 * 3.33e-3  # microns
lon_range = np.arange(2.7, 5.3, dx)   # microns
lon_mid_y = 4.                                   # microns

myCmaps = {}
colors = {"H": "blue", "D": "darkorange", "e": "red"}

# generate the colors for colormaps
def generateCmaps():
    for s in colors.keys():
        myColor1 = colorConverter.to_rgba('white', alpha=1.0)
        myColor2 = colorConverter.to_rgba(colors[s], alpha=1.0)
        myCmap = mpl.colors.LinearSegmentedColormap.from_list(
            'my_cmap_' + s, [myColor1,myColor2], 1200)
        myCmap._init() # create the _lut array, with rgba values
        myCmaps[s] = myCmap

def getData(species, sfilter, a0, step, geom, dratio, suffix=""):
    """..."""
    h5f = h5.File(
        "a0_" + str(a0) +
        "/2D_" + geom + "_dratio" + dratio + suffix +
        "/simOutput/h5/ions_" + str(step) + ".h5",
        "r")

    h5species = h5f["/data/{0}/particles/{1}_{2}".format(str(step), species, sfilter)]

    # data
    d_y0 = h5species["y0"]
    d_w = h5species["weighting"]
    d_mom_x = h5species["momentum/x"]
    d_mom_y = h5species["momentum/y"]
    d_mom_z = h5species["momentum/z"]
    d_mass = h5species["mass"]

    y0_mu = d_y0[()] * d_y0.attrs["unitSI"] * 1.e6
    w = d_w[()] * d_w.attrs["unitSI"]
    mom_x = d_mom_x[()] / w * d_mom_x.attrs["unitSI"]
    mom_y = d_mom_y[()] / w * d_mom_y.attrs["unitSI"]
    mom_z = d_mom_z[()] / w * d_mom_z.attrs["unitSI"]
    direc = np.ones_like(mom_y)
    direc[mom_y < 0.] = -1.0
    mass = d_mass.attrs["value"] * d_mass.attrs["unitSI"]

    # hack: fw only
    # w[mom_y<0.] = 0.

    h5f.close()

    # constants
    c2 = 2.99792458e8**2
    m2_c2_reci = 1. / (mass * mass * c2)
    J2MeV = 6.242e+12

    # energy [MeV] / A
    mom2 = mom_x*mom_x + mom_y*mom_y + mom_z*mom_z
    gamma = np.sqrt(1.0 + mom2 * m2_c2_reci)
    ene_MeV = (gamma - 1.0) * mass* c2 * J2MeV * direc

    if species == 'D':
        ene_MeV = ene_MeV * 0.5

    if sfilter.startswith('fw'):
        ene_range = np.linspace(0.0, max_energy[a0], Nene)  # MeV
    else:
        ene_range = np.linspace(-max_energy[a0], max_energy[a0], Nene)  # MeV

    # histogram
    counts, y_edges, ene_edges = np.histogram2d(
        y0_mu,
        ene_MeV,
        bins=(
            lon_range,
            ene_range
        ),
        weights=w
    )
    return counts.T, y_edges, ene_edges

def plotAndSave(species, sfilter, a0, step, geom, dratios, suffix=""):
    """..."""
    for dratio in dratios:
        if species == "H" and dratio == "1.00":
            continue
        if species == "D" and dratio == "0.00":
            continue

        f = plt.figure(figsize=[4, 3])

        # left, bottom, W, H
        ax = f.add_axes([0.15, 0.15, 0.50, 0.75])
        cax = f.add_axes([0.87, 0.15, 0.03, 0.75])
        ax_ene = f.add_axes([0.69, 0.15, 0.15, 0.75], sharey=ax)

        # target
        ax.axvline(x=lon_mid_y - 1.,
            zorder=1,
            linestyle='-',
            linewidth=0.8,
            color='black')
        ax.axvline(x=lon_mid_y + 1.,
            zorder=1,
            linestyle='-',
            linewidth=0.8,
            color='black')
        ax.axvline(x=lon_mid_y,
            zorder=1,
            linestyle=':',
            linewidth=0.6,
            color='black')

        counts, y_edges, ene_edges = \
            getData(species, sfilter, a0, step, geom, dratio, suffix)
        extent = [y_edges[0], y_edges[-1], ene_edges[0], ene_edges[-1]]

        im = ax.imshow(
            counts,
            extent = extent,
            interpolation = 'nearest',
            aspect = 'auto',
            origin='lower',
            norm = LogNorm(),
            cmap=myCmaps[species],
            zorder=3,
            vmin=1.e1,
            vmax=1.e6
        )

        cbar = plt.colorbar(im, cax=cax)
        cbar.ax.set_title('count\n[arb. u.]',
            loc='left', size=ax.yaxis.label.get_fontsize() * 0.75)

        lns_for_legend = []

        # deterium cutoff energy
        dratiof = float(dratio)
        if species == 'H' and dratiof > 0.0:
            cutoff_name = geom + '/energyCutoff/cutoffs_' + sfilter
            with open(cutoff_name + '.json', 'r') as fp:
                cutoff_energies = json.load(fp)
            D_cutoff_energy = cutoff_energies[geom][str(a0)]['D'][str(dratiof)]

            lns_for_legend.append(ax_ene.axhline(y=D_cutoff_energy,
                color=colors['D'],
                zorder=2,
                label="D: cutoff"))
            ax.axhline(y=D_cutoff_energy,
                color=colors['D'],
                zorder=2,
                label="D: cutoff")

        # energy histogram
        #   front/rear
        idx_mid_y = next(x[0] for x in enumerate(lon_range) if x[1] > lon_mid_y)
        ax_ene_color = "red"
        ax_ene.plot(
            np.sum(counts, axis=1),
            ene_edges[:-1],
            color=ax_ene_color,
            linewidth=1.7,
            label="sum"
        )
        lns_for_legend.extend(ax_ene.plot(
            np.sum(counts[:,:idx_mid_y], axis=1),
            ene_edges[:-1],
            color='lightgreen',
            linewidth=1.0,
            linestyle='-',
            label="front"
        ))
        lns_for_legend.extend(ax_ene.plot(
            np.sum(counts[:,idx_mid_y:], axis=1),
            ene_edges[:-1],
            color=colors[species],
            linewidth=1.0,
            linestyle='-',
            label="rear"
        ))
        ax_ene.text(s="sum\n{0}".format(sfilter),
                color=ax_ene_color,
                ha='center',
                size=ax.yaxis.label.get_fontsize() * 0.75,
                x=1.e4, y=max_energy[a0] * 0.9)
        ax.text(s="$\mathregular{{a}_0}$ = " + "{0}".format(a0),
                color=ax_ene_color,
                ha='center',
                size=ax.yaxis.label.get_fontsize() * 0.85,
                x=lon_mid_y, y=max_energy[a0] * 0.6)
        ax_ene.set_xlim([1.e1, 1.e7])
        #ax_ene.set_position(ax.get_position())
        ax_ene.set_xscale("log")
        #ax_ene.spines['top'].set_color(ax_ene_color)
        #ax_ene.tick_params(axis='x', which='both', colors=ax_ene_color)
        #ax_ene.xaxis.label.set_color(ax_ene_color)
        ax_ene.tick_params(labelleft=False)
        labs = [l.get_label() for l in lns_for_legend]
        ax.legend(lns_for_legend, labs, loc='upper center')

        # initial density
        #plt.plot(
        #    lon_range[:-1],
        #    np.sum(counts, axis=0)
        #)

        # limits
        ene_org_lim = ax_ene.get_xlim()
        ax_ene.set_xlim([ene_org_lim[1], ene_org_lim[0]])
        ax.set_xlim([lon_range[0], lon_range[-1]])
        ax.set_ylim([ene_edges[0], ene_edges[-1]])

        # annotations
        ax.set_xlabel(r'y$_0$  [$\mu$m]')
        ax.set_ylabel(r'Ene / A  [MeV]')

        # ax.set_title(
        #     "{0} ({1}) @ t = {2}".format("-".join(species), sfilter, step))
        #plt.legend()

        #plt.show()
        plt.savefig(geom + "/energyOrigin/" +
            "a0_" + str(a0) + "_" + species + "_" + sfilter + "_" +
            geom +
            "_dratio" + dratio +
            "_energyOrigin.svg", dpi=300)
        plt.close()

generateCmaps()

#for species in ['H', 'D']:
#    plotAndSave(species, sfilter, a0, step[a0], geom, dratios=["0.00", "0.10", "0.50", "0.90", "0.98", "1.00"])

#            mixed_30nc separ_30nc  separ_300nc separ_1000nc separ_1700nc
for geom in ['mixed_30nc']:  # 'separ_30nc'
    for a0 in [16, 30, 42]:
        for species in ['H', 'D']:
            for sfilter in ['fwPinhole', 'fwRCF']:
                plotAndSave(species, sfilter, a0, step[a0], geom, dratios=["0.00", "0.10", "0.50", "0.60", "0.70", "0.80", "0.90", "0.98", "1.00"])
