#!/usr/bin/env python
#
# Copyright 2018-2019 Axel Huebl
#
# License: CC-BY-SA 4.0 International
#
import matplotlib as mpl
#mpl.use('Agg') # non-interactive backend
import matplotlib.pyplot as plt
from matplotlib.colors import colorConverter, LogNorm, Normalize
import numpy as np
import pandas as pd
import h5py as h5
import pickle
from scipy.optimize import curve_fit
from os import sys, path


all_dratios = [
    "0.00", "0.10", "0.20", "0.30", "0.40", "0.50",
    "0.60", "0.70", "0.80", "0.90", "0.98", "1.00"
]

# relevant dratios for plotted comparison of densities & fields
#   select lines that show relevant changes
lines_select = [
    "0.00",
    # "0.10",
    "0.50",
    #"0.70",
    "0.80",
    # "0.90",
    "0.98",
    # "1.00"
]
# lines_select = all_dratios

lambda_0 = 800.e-9  # m
n0 = 30.            # n_c
mu = 1.e6
nc = 1.11485e21 * 1.e6 / (lambda_0 * mu) / (lambda_0 * mu)
y_rear = 5.0        # behind target rear, in microns
y_low  = 3.1  # 3.1, y_rear
y_hig  = 4.9

smooth_long = 3  # cells in y (1=3.33nm, 20=0.066 microns), only E_y

myCmaps = {}
colors = {"H": "blue", "D": "darkorange", "e": "red", "E": "purple"}

# see https://stackoverflow.com/a/8702435/2719194
#     https://stackoverflow.com/a/26497091/2719194
def nested_dict(): return defaultdict(nested_dict)

def ne_0(y):
    """y in microns, returned in n_c"""
    s = np.ones_like(y)

    d = 2.0   # lateral (longitudinal) size
    L = 0.02  # exponential "fuzzyness" factor around the rectangle

    mid_y = 4.0  # lateral (longitudinal) middle

    # ramps
    abs_y = np.abs( y * -1.0 + mid_y )
    targetBorder_y = d/2.0 - L

    in_target = abs_y > targetBorder_y
    s[in_target] = np.exp( (abs_y[in_target] * -1.0 + targetBorder_y) / L)

    # hard cutoff after 6*L
    s[abs_y > targetBorder_y + 6.0 * L] = 0.0

    return s * n0;

def getField(a0, step, geom, dratio, field, component=None, x_pm=0.05):
    """..."""
    import h5py as h5
    sim_dir = "a0_" + str(a0) + "/2D_" + geom + "_dratio" + dratio + "/"
    f = h5.File(sim_dir + "simOutput/h5/fields_" + str(step) + ".h5", "r")

    record = f["/data/" + str(step) + "/fields/" + field]
    if component is not None:
        rc = record[component]
    else:
        rc = record

    shape = rc.shape
    unit = rc.attrs["unitSI"]

    dy_unit = record.attrs["gridUnitSI"]
    dy = record.attrs["gridSpacing"][0]
    y_off = record.attrs["gridGlobalOffset"][0]
    y_bins = np.arange(shape[0]) * ( dy * dy_unit )
    y_bins = y_bins + y_off * dy_unit

    # single line sliced out
    #          y, x
    # return rc[:, shape[1]//2], unit

    # sfilter == 'onLaserAxis' means
    #   0.45 to 0.55 around center axis
    # x_pm = 0.5  # 0.05 0.15 0.5
    x_min = int(shape[1] * (0.5 - x_pm))
    x_max = int(shape[1] * (0.5 + x_pm))
    rc_slice = rc[:, x_min:x_max]
    return np.average(rc_slice, axis=1), y_bins, unit

def getCutoffData(sfilter, geom):
    save_name = geom + '/energyCutoff/cutoffs_' + sfilter
    with open(save_name + '.p', 'rb') as fp:
        save_cutoff = pickle.load(fp)
    return save_cutoff

def plotScalesOverCutoff(a0, step, geom, dratios):
    f = plt.figure(figsize=[4, 2.8])

    # left, bottom, W, H
    ax_cutoff = f.add_axes([0.2, 0.15, 0.57, 0.375])
    ax_NeT = ax_cutoff.twinx()
    ax_Ne = f.add_axes([0.2, 0.525, 0.57, 0.375], sharex=ax_cutoff)
    ax_T = ax_Ne.twinx()

    # data
    save_cutoff = getCutoffData(sfilter='fwPinhole', geom=geom)
    d_sorted = sorted(save_cutoff[geom][a0]['H'].keys())
    cutoffv = [save_cutoff[geom][a0]['H'][d] for d in d_sorted]

    # fields
    Ne_rear = {}      # count
    NeT_rear = {}     # energy
    for dratio in dratios:
        ne, y_bins, ne_unit = getField(a0, step, geom, dratio, 'e_density', None, 0.1)
        neT, _, neT_unit = getField(a0, step, geom, dratio, 'e_energyDensity', None, 0.1)

        #rear_min = np.argmax(y_bins * mu > y_rear)  # electrons behind rear
        y_idx_low = np.argmax(y_bins * mu > y_low)
        y_idx_hig = np.argmax(y_bins * mu > y_hig)
        Ne_rear[dratio] = np.abs(np.mean(ne[y_idx_low:y_idx_hig]) * ne_unit)
        NeT_rear[dratio] = np.abs(np.mean(neT[y_idx_low:y_idx_hig]) * neT_unit)

    d_list = [float(d) for d in sorted(Ne_rear.keys())]
    Ne = [Ne_rear[d] for d in sorted(Ne_rear.keys())]
    # Ne = [Ne_front[d] for d in sorted(Ne_front.keys())]
    NeT = [NeT_rear[d] for d in sorted(NeT_rear.keys())]
    Ne = np.array(Ne)
    NeT = np.array(NeT)
    T = NeT / Ne

    # L_front = [L_front_nm[d] for d in sorted(L_front_nm.keys())]

    MeV2J = 1.60218e-13  # J
    J2MeV = 1. / MeV2J   # MeV / J

    # plot
    p_cutoff, = ax_cutoff.plot(
        d_sorted,
        cutoffv,
        color='grey',
        marker='x',
        label=r'max. Ene / A')
    p_NeT, = ax_NeT.plot(
        d_list,
        NeT  / nc * J2MeV,
        marker='o',
        ls="--",
        color='red',
        label=r'$<T_e> \cdot N_e$')
        #label=r'$\int_{5\,\mathrm{\mu m}}^{\infty} T\,n_e(y) \,\mathrm{d}y$')
    p_Ne, = ax_Ne.plot(
        d_list,
        np.array(Ne) / nc / n0,
        marker='o',
        ls="--",
        color='blue',
        label=r'$N_e$')
        #label=r'$\int_{5\,\mathrm{\mu m}}^{\infty} n_e(y) \,\mathrm{d}y$')
        #label=r'$\int_{3\,\mathrm{\mu m}}^{y(n_0)} n_e(y) \,\mathrm{d}y$')
    p_T, = ax_T.plot(
        d_list,
        T * J2MeV,
        marker='x',
        color='black',
        label=r'$<T_e>$')
        #label=r'$\int_{5\,\mathrm{\mu m}}^{\infty} T \,\mathrm{d}y$')

    # annotate
    #ax_cutoff.legend(handles=[p_cutoff, p_NeT])
    #ax_T.legend(handles=[p_T, p_Ne])

    ax_Ne.tick_params(labelbottom=False)
    ax_NeT.tick_params(labelbottom=False)

    ax_cutoff.spines['left'].set_color('grey')
    ax_cutoff.yaxis.label.set_color('grey')
    ax_cutoff.tick_params(axis='y', colors='grey')
    ax_Ne.spines['left'].set_color('blue')
    ax_Ne.yaxis.label.set_color('blue')
    ax_Ne.tick_params(axis='y', colors='blue')

    ax_NeT.spines['right'].set_color('red')
    ax_NeT.yaxis.label.set_color('red')
    ax_NeT.tick_params(axis='y', colors='red')
    ax_T.spines['right'].set_color('black')
    ax_T.yaxis.label.set_color('black')
    ax_T.tick_params(axis='y', colors='black')

    ax_cutoff.set_xlabel('Deuterium ratio $d$')
    ax_cutoff.set_ylabel('max. Ene / A\n[MeV]')
    ax_NeT.set_ylabel('$<T_e> \cdot N_e$\n[MeV $n_c$]')
    ax_T.set_ylabel('$<T_e>$\n[MeV]')
    ax_Ne.set_ylabel('$n_e$\n[$n_0$]')
    #ax_NeT.set_ylabel('$\int_{5\,\mathrm{\mu m}}^{\infty} T\,n_e$\n[arb.u.]')
    #ax_T.set_ylabel('$\int_{5\,\mathrm{\mu m}}^{\infty} T$\n[arb.u.]')
    #ax_Ne.set_ylabel('$\int_{5\,\mathrm{\mu m}}^{\infty} n_e$\n[arb.u.]')

    ax_Ne.set_title(r'$a_0 = {0}$'.format(a0), color='red')

    fname = (geom + "/fieldsRear/a0_" +
        str(a0) + "_onLaserAxis_" +
        geom +
        "_frontRearScales_")
    plt.savefig(fname + str(step) + ".svg", dpi=300)
    plt.close()

def plotAndSave(a0, step, geom, dratios):
    """..."""
    Ne_front = {}     # count
    Ne_rear = {}      # count
    NeT_rear = {}     # energy
    L_front_nm = {}   # front scale length in nm
    for dratio in dratios:
        f = plt.figure(figsize=[4, 3])

        # left, bottom, W, H
        # ax_dens = f.add_axes([0.15, 0.15, 0.70, 0.75])
        ax_dens = f.add_axes([0.18, 0.17, 0.70, 0.7])
        ax_Ey = ax_dens.twinx()

        ne, _, ne_unit = getField(a0, step, geom, dratio, 'e_density')
        nH, _, nH_unit = getField(a0, step, geom, dratio, 'H_density')
        nD, _, nD_unit = getField(a0, step, geom, dratio, 'D_density')
        neT, _, neT_unit = getField(a0, step, geom, dratio, 'e_energyDensity')
        E_y, y_bins, Ey_unit = getField(a0, step, geom, dratio, 'E', 'y')

        # unit conversion from SI
        mu = 1.e6  # meters to microns
        #e_mc_r = 1. / (9.109e-31 * 2.9979e8)  # electrons: kg * m / s to beta * gamma
        #H_mc_r = e_mc_r / 1836.152672         # protons: kg * m / s to beta * gamma
        #D_mc_r = H_mc_r / 1.9985              # protons: kg * m / s to beta * gamma

        # relativistic density ne / gamma
        me_c2 = 8.1871e-14  # J (=511 keV)
        avg_ene_e = neT / ne * (neT_unit / ne_unit)
        avg_gamma_e = avg_ene_e / me_c2 + 1.0

        # plotting: initial density
        #ax_dens.axhline(n0,
        #    alpha=0.1,
        #    lw=1,
        #    color='grey')
        #ax_dens.axvline(y_rear - 2.0,
        #    alpha=0.1,
        #    lw=1,
        #    color='grey')
        #ax_dens.axvline(y_rear,
        #    alpha=0.1,
        #    lw=1,
        #    color='grey')
        ax_dens.plot(y_bins * mu, ne_0(y_bins * mu),
            label=r'$n_{e,0}$',
            lw=1,
            color='grey'
        )
        # plotting: relativistic n_c
        #   all n_e below? transparent! at all times n_e > gamma n_c? opaque!
        ax_dens.plot(y_bins * mu, avg_gamma_e,
            color='black',
            label=r'$\overline{\gamma}\,n_c$'
        )

        # plotting: density
        if dratio != "0.00":
            ax_dens.plot(y_bins * mu, nD * nD_unit / nc,
                color=colors['D'],
                label=r'$n_D$'
            )
        if dratio != "1.00":
            ax_dens.plot(y_bins * mu, nH * nH_unit / nc,
                color=colors['H'],
                label=r'$n_H$'
            )
        ax_dens.plot(y_bins * mu, ne * ne_unit / nc,
            color=colors['e'],
            label=r'$n_e$'
        )
        #ne_igamma = ne / avg_gamma_e
        #ax_dens.plot(y_bins * mu, ne_igamma * ne_unit / nc,
        #    color='black',
        #    label=r'$n_e / \gamma$'
        #)

        # analyze: difference to initial electron density
        d_ne = ne_0(y_bins * mu) - ne * ne_unit / nc
        #fw_displaced = np.argmax(ne * ne_unit / nc > n0)
        #end_displaced = np.argmax(y_bins * mu > 4.5)  # for step = 10'000
        rear_min = np.argmax(y_bins * mu > y_rear)  # electrons behind rear
        #print("rear_min = {0} ({1})".format(y_bins[rear_min] * mu, rear_min))
        #Ne = np.sum(d_ne[0:rear_min])
        #Ne = np.sum(d_ne[rear_min:end_displaced])
        Ne = np.abs(np.sum(d_ne[rear_min:]))
        print("Ne displaced rear: a0={0} step={1} d={2} Ne={3}".format(a0, step, dratio, Ne))
        Ne_rear[dratio] = Ne

        front_min = np.argmax(y_bins * mu > y_rear - 2.0)
        front_max = np.argmax(ne * ne_unit / nc > n0)
        Ne = np.abs(np.sum(d_ne[front_min:front_max]))
        Ne_front[dratio] = Ne

        # electron energy on rear
        NeT = np.abs(np.sum(neT[rear_min:]) * neT_unit)
        NeT_rear[dratio] = NeT

        # analyse: scale length on target front
        ni = ( nH * nH_unit + nD * nD_unit ) / nc
        n_min = 1.0
        n_max = (1.+0.5*a0**2)**(0.5)-1.  # pond; 10.0
        b_min = np.argmax(ni > n_min)
        b_max = np.argmax(ni > n_max)
        n_min = ni[b_min]
        n_max = ni[b_max]
        y_min = y_bins[b_min] * mu
        y_max = y_bins[b_max] * mu
        print(y_min, y_max, n_min, n_max)
        y_fits = y_bins[b_min:b_max] * mu
        #fit_front = lambda y_fits, L: n_max * np.exp((y_fits-y_max) / L)
        fit_front = lambda y_fits, L: np.log(n_max) + (y_fits-y_max) / L
        try:
            popt, pcov = curve_fit(
                fit_front,
                y_fits,
                #ni[b_min:b_max],
                np.log(ni[b_min:b_max]),
                bounds=([0.], [np.inf])
            )

            L_nm = tuple(popt)[0] * 1.e3
        except:
            print("front fit: fit failed!")
            L_nm = 0.001
        L_front_nm[dratio] = L_nm
        print("front fit (a0={0}, d={1}): L={2:.2f}nm".format(a0, dratio, L_nm ))
        #ax_dens.plot(
        #    y_fits,
        #    # fit_front(y_fits, L_nm * 1.e-3),
        #    np.exp(fit_front(y_fits, L_nm * 1.e-3)),
        #    lw=1,
        #    ls="--",
        #    color="black",
        #    label="front fit")

        # plotting: electric field
        dy = y_bins[1] - y_bins[0]
        cells_per_lambda_half = int(round(lambda_0 / 2.0 / dy))  # ~240
        #   smooth histogram by merging neighboring bins
        #   counts, bins = merge_bins(E_y * E_unit, y, cells_per_lambda_half)
        df = pd.DataFrame({'E' : E_y * Ey_unit, 'y' : y_bins})
        # r = df.rolling(window=cells_per_lambda_half, win_type='blackman')
        r = df.rolling(window=smooth_long, win_type='blackman')

        #ax_Ey.plot(
        #    y_bins * mu,
        #    E_y * Ey_unit * 1.e-12,
        #    label=r'$E_{||}$',
        #    lw=1,
        #    alpha=0.5,
        #    color='grey')
        ax_Ey.plot(
            r['y'].mean() * mu,
            r['E'].mean() * 1.e-12,
            label=r'$E_{||}$',
            lw=1.5,
            alpha=0.7,
            color=colors['E'])
        #ax_Ey.plot(
        #    r['y'].mean() * mu,
        #    r['E'].mean() * 1.e-12,
        #    label=r'$<E_{||}>^\mathrm{blackman}_{\lambda_0 / 2}$',
        #    lw=1,
        #    alpha=0.7,
        #    color=colors['E'])

        # limits
        if step <= 34000:
            ax_dens.set_xlim([2., 11.])
        elif step <= 50000:
            ax_dens.set_xlim([2., 19.])
        # ax_Ey.set_ylim([0., None])
        #ax_Ey.set_ylim([0., 14.0])
        #ax_Ey.set_ylim([0., 10.0])
        if step <= 10000:
            ax_Ey.set_ylim([0., 14.0])
        else:
            ax_Ey.set_ylim([0., 4.0])
            # ax_Ey.set_ylim([0., 10.0])
        # ax_dens.set_ylim([0.01, 70.])
        ax_dens.set_ylim([0.01, 100.])
        ax_dens.set_yscale('log')

        # annotations
        ax_dens.set_xlabel(r'$y$ [$\mathrm{\mu m}$]')
        # FIXME: is this scaling useful for multi-ion compares?!
        ax_dens.set_ylabel(r'$n$ [$n_c$]')

        #ax_Ey.tick_params(axis='y',direction='in', pad=-25)
        ax_Ey.spines['right'].set_color(colors['E'])
        ax_Ey.yaxis.label.set_color(colors['E'])
        ax_Ey.tick_params(axis='y', colors=colors['E'])
        ax_Ey.set_ylabel(r'$E_y$  [TV / m]')

        plt.title("H & {0} D ({1}) @ t = {2}".format(dratio, "onLaserAxis", step))
        ax_dens.legend(loc='lower right')
        ax_Ey.legend(loc='upper right')

        #plt.show()
        fname = (geom + "/fieldsRear/a0_" +
            str(a0) + "_H_D_onLaserAxis_" +
            geom +
            "_dratio" + dratio +
            "_rearFields_")
        plt.savefig(fname + str(step) + ".svg", dpi=300)
        plt.close()

    # plotScalesOverCutoff(a0, step, geom, all_dratios)  # L_front_nm


for geom in ['mixed_30nc']:
    for a0 in [8, 16, 23, 30, 42]:
        #for step in range(10000, 15001, 5000):
        for step in range(10000, 25001, 5000):
            print(a0, step, geom)
            plotAndSave(a0, step, geom, dratios=lines_select)

    for a0 in [8, 16, 23, 30, 42]:
        #plotScalesOverCutoff(a0,  5000, geom, all_dratios)  # -25fs before peak @8190
        plotScalesOverCutoff(a0, 10000, geom, all_dratios)  #  14fs after peak @8190
        #plotScalesOverCutoff(a0, 15000, geom, all_dratios)  #  53fs after peak @8190
        #plotScalesOverCutoff(a0, 20000, geom, all_dratios)  #  93fs after peak @8190
        #plotScalesOverCutoff(a0, 25000, geom, all_dratios)  # 132fs after peak @8190
