#!/usr/bin/env python
#
# Copyright 2018-2019 Axel Huebl
#
# License: CC-BY-SA 4.0 International
#
import matplotlib as mpl
#mpl.use('Agg') # non-interactive backend
import matplotlib.pyplot as plt
from matplotlib.colors import colorConverter, LogNorm, Normalize
from picongpu.plugins.data import PhaseSpaceData
import numpy as np
import pandas as pd
from os import sys, path


# a0=16  # 16, 30, 42

# sfilter='fwRCF'        # 'onLaserAxis', 'fwRCF', 'fwPinhole'
geom='mixed_30nc'      # mixed_30nc separ_30nc separ_300nc separ_1000nc separ_1700nc

myCmaps = {}
colors = {"H": "blue", "D": "darkorange", "e": "red"}

def merge_bins(counts, bins, num_merge):
    """
    Merge num_merge bins into one to smooth histograms.
    Normalize count by same factor.
    """
    new_counts = np.zeros(counts.shape[0] // int(num_merge), dtype=counts.dtype)
    new_bins = np.zeros(bins.shape[0] // int(num_merge), dtype=bins.dtype)

    new_bins[:] = bins[::num_merge]
    for i in np.arange(num_merge):
        new_counts[:] += counts[i::num_merge]

    new_counts = new_counts / float(num_merge)

    return new_counts, new_bins

# generate the colors for colormaps
def generateCmaps():
    for s in colors.keys():
        myColor1 = colorConverter.to_rgba('white', alpha=1.0)
        myColor2 = colorConverter.to_rgba(colors[s], alpha=1.0)
        myCmap = mpl.colors.LinearSegmentedColormap.from_list(
            'my_cmap_' + s, [myColor1,myColor2], 1200)
        myCmap._init() # create the _lut array, with rgba values
        myCmaps[s] = myCmap

def getData(species, sfilter, a0, step, geom, dratio, ps_name='ypy'):
    """..."""
    phase_space = PhaseSpaceData("a0_" + str(a0) + "/2D_" + geom + "_dratio" + dratio)
    ps, ps_meta = phase_space.get(species, sfilter, iteration=step, ps=ps_name)

    # convert to momentum per nukleon
    #if species == 'D':
    #    ps_meta.extent = ps_meta.extent * [1., 1., 0.5, 0.5]

    return ps, ps_meta

def getEfield(a0, step, geom, dratio, component='y'):
    """..."""
    import h5py as h5
    sim_dir = "a0_" + str(a0) + "/2D_" + geom + "_dratio" + dratio + "/"
    f = h5.File(sim_dir + "simOutput/h5/fields_" + str(step) + ".h5", "r")

    E_y = f["/data/" + str(step) + "/fields/E/" + component]
    s = E_y.shape

    # single line sliced out
    #          y, x
    # return E_y[:, E_y.shape[1]//2], E_y.attrs["unitSI"]

    # sfilter == 'onLaserAxis' means
    #   0.45 to 0.55 around center axis
    x_min = int(E_y.shape[1] * 0.45)
    x_max = int(E_y.shape[1] * 0.55)
    E_y_slice = E_y[:, x_min:x_max]
    return np.average(E_y_slice, axis=1), E_y.attrs["unitSI"]

def plotAndSave(sfilter, a0, step, geom, dratios):
    """..."""
    for dratio in dratios:
        f = plt.figure(figsize=[4, 3])
        # left, bottom, W, H
        ax = f.add_axes([0.18, 0.17, 0.55, 0.7])
        cax_H = f.add_axes([0.74, 0.17, 0.03, 0.7])
        cax_D = f.add_axes([0.79, 0.17, 0.03, 0.7])
        # ax = plt.gca()

        #e_ps, e_ps_meta = getData('e', sfilter, a0, step, geom, dratio)
        H_ps, H_ps_meta = getData('H', sfilter, a0, step, geom, dratio)
        D_ps, D_ps_meta = getData('D', sfilter, a0, step, geom, dratio)

        # unit conversion from SI
        mu = 1.e6  # meters to microns
        e_mc_r = 1. / (9.109e-31 * 2.9979e8)  # electrons: kg * m / s to beta * gamma
        H_mc_r = e_mc_r / 1836.152672         # protons: kg * m / s to beta * gamma
        D_mc_r = H_mc_r / 1.9985              # protons: kg * m / s to beta * gamma

        # plotting
        ax.axhline(0.0,
            alpha=0.1,
            lw=1,
            color='black')
        ax.axvline(3.0,
            alpha=0.1,
            lw=1,
            color='grey')
        ax.axvline(5.0,
            alpha=0.1,
            lw=1,
            color='grey')
        if dratio != "0.00":
            im_D = ax.imshow(
                np.abs(D_ps).T * D_ps_meta.dV,
                extent = D_ps_meta.extent * [mu, mu, D_mc_r, D_mc_r],
                interpolation = 'nearest',
                aspect = 'auto',
                origin='lower',
                norm = LogNorm(),
                cmap=myCmaps['D'],
                label='D',
                vmin=1.e-28,
                vmax=1.e-23
            )
        if dratio != "1.00":
            im_H = ax.imshow(
                np.abs(H_ps).T * H_ps_meta.dV,
                extent = H_ps_meta.extent * [mu, mu, H_mc_r, H_mc_r],
                interpolation = 'nearest',
                aspect = 'auto',
                origin='lower',
                norm = LogNorm(),
                cmap=myCmaps['H'],
                label='H',
                vmin=1.e-28,
                vmax=1.e-23
            )

        cbar_H = plt.colorbar(im_H, cax=cax_H)
        if dratio != "0.00":
            cbar_D = plt.colorbar(im_D, cax=cax_D)
            cbar_D.set_label(r'$Q / \mathrm{d}r \mathrm{d}p$ [$\mathrm{C s kg^{-1} m^{-2}}$]')
        cbar_H.ax.set_yticklabels([])
        # cbar_H.set_label(r'$Q / \mathrm{d}r \mathrm{d}p$ [$\mathrm{C s kg^{-1} m^{-2}}$]')

        # limits
        if step <= 15000:
            ax.set_xlim([2.8, 7.5])
        elif step <= 20000:
            ax.set_xlim([2.8, 9.5])
        elif step <= 25000:
            ax.set_xlim([2.8, 11.])

        #if step <= 34000:
        #    ax.set_xlim([2.4, 11.])
        #elif step <= 50000:
        #    ax.set_xlim([0., 19.])

        # fw filter only? py >= 0
        if sfilter.startswith("fw"):
            p_ax_range = list(ax.get_ylim())
            p_ax_range[0] = 0.0
            p_ax_range[1] *= 0.6
            ax.set_ylim(p_ax_range)

        # annotations
        ax.set_xlabel(r'${0}$'.format(H_ps_meta.r) + r' [$\mathrm{\mu m}$]')
        # FIXME: is this scaling useful for multi-ion compares?!
        ax.set_ylabel(r'$p_{0}$ [$\beta\gamma$]'.format(H_ps_meta.p))

        f.text(s='H',
            color=colors['H'],
            x=0.68, y=0.25)
        f.text(s='D',
            color=colors['D'],
            x=0.68, y=0.20)

        #ax.title("H & {0} D ({1}) @ t = {2}".format(dratio, sfilter, step))
        #ax.legend()

        #plt.show()
        fname = (geom + "/ps_ypy/a0_" + str(a0) + "/a0_" +
            str(a0) + "_H_D_" + sfilter + "_" +
            geom +
            "_dratio" + dratio +
            "_ypy_")
        f.savefig(fname + str(step) + ".svg", dpi=300)
        plt.close()

        #return
        continue

        # get E field if available
        try:
            E_y, E_unit = getEfield(a0, step, geom, dratio, 'y')
            E_ax = ax.twinx()

            dy = H_ps_meta.dV**(1./3.)
            y = np.arange(0, E_y.shape[0]) * dy * 1.e6
            lambda_0 = 800.e-9  # microns
            cells_per_lambda_half = int(round(lambda_0 / 2.0 / dy))
            # smooth histogram by merging neighboring bins
            # counts, bins = merge_bins(E_y * E_unit, y, cells_per_lambda_half)
            df = pd.DataFrame({'E' : E_y * E_unit, 'y' : y})
            # r = df.rolling(window=cells_per_lambda_half, win_type='blackman')
            r = df.rolling(window=3, win_type='blackman')

            #E_ax.plot(
            #    y,
            #    E_y * E_unit,
            #    label=r'$E_{||}$',
            #    lw=1,
            #    alpha=0.5,
            #    color='grey')
            E_ax.plot(
                r['y'].mean(),
                r['E'].mean(),
                label=r'$E_{||}$',
                lw=1,
                alpha=0.5,
                color='purple')
            #E_ax.plot(
            #    r['y'].mean(),
            #    r['E'].mean(),
            #    label=r'$<E_{||}>^\mathrm{blackman}_{\lambda_0 / 2}$',
            #    lw=1,
            #    alpha=0.7,
            #    color='purple')

            minmax = np.max([-np.min(r['E'].mean()), np.max(r['E'].mean())])

            # common zero for both y axis
            p_ax_range = ax.get_ylim()
            dp_to_mid = p_ax_range[1] / (p_ax_range[1] - p_ax_range[0])

            E_ax.set_ylim([
                -3. * minmax * ( 1. - dp_to_mid ),
                 3. * minmax * dp_to_mid])
            # E_ax.set_ylim([-2. * minmax, 2. * minmax])

            E_ax.tick_params(axis='y',direction='in', pad=-25)
            E_ax.spines['right'].set_color('purple')
            E_ax.yaxis.label.set_color('purple')
            E_ax.tick_params(axis='y', colors='purple')
            E_ax.set_ylabel(r'$E_y$  [V / m]')

            plt.legend(loc='upper left')
            plt.savefig(fname + "Ey_" + str(step) + ".svg", dpi=300)
        except:
            pass

        plt.close()


generateCmaps()

#all_dratios = [
#    "0.00", "0.10", "0.20", "0.30", "0.40", "0.50",
#    "0.60", "0.70", "0.80", "0.90", "0.98", "1.00"
#]
all_dratios = ["0.00", "0.80", "0.90"]

# step 30'000 is 260fs after peak on target
for geom in ['mixed_30nc']:  # 'separ_30nc'
    for a0 in [16, 42]:  # [8, 16, 23, 30, 42]:
        for step in range(5000, 25001, 5000):
        # for step in range(5000, 50001, 5000):
        # for step in range(5000, 50001, 1000):
            for sfilter in ['fwPinhole']:  #, 'onLaserAxis', 'fwRCF']:
                print(sfilter, a0, step, geom)
                plotAndSave(sfilter, a0, step, geom, dratios=all_dratios)
